import builtins
import types
from typing import Iterable
from i2 import Sig
from i2.signatures import _remove_variadics_from_sig
from i2.tests.util import sig_to_inputs, _sig_to_inputs, sig_to_func
from i2.signatures import VP, VK


def get_builtin_function_types():
    builtin_function_names = [
        name
        for name, obj in vars(builtins).items()
        if isinstance(obj, types.BuiltinFunctionType)
    ]
    return builtin_function_names


def get_lines_from_string(text):
    return text.split('\n')


def extract_doc(doc):
    lines = get_lines_from_string(doc)
    result = []  # better: yield and do list()
    for line in lines:
        if line.strip():
            result.append(line)
        else:
            break
    return result


def enclosed_text(text):
    idx1 = text.find('(')
    idx2 = text.find(')', idx1)

    if idx1 != -1 and idx2 != -1 and (idx2 > idx1 + 1):
        result = text[idx1 + 1 : idx2]
    return result


def args_for_builtin(builtin_func, prepare_for_Sig=False):
    """Extract the args from builtin doc"""

    docstr = builtin_func.__doc__
    doc = extract_doc(docstr)
    name = builtin_func.__name__

    result = []
    for line in doc:
        s = line.replace(name, '', 1)
        args = enclosed_text(s)
        if prepare_for_Sig:
            args = args.replace(',', '')
        result.append(args)

    return result


def args_str_from_sig(sig):
    return str(sig)[1:-1]


def sig_to_lambda(sig):
    args = args_str_from_sig(sig)
    lambda_func = eval(f'lambda {args}:  None')
    return lambda_func


def builtin_func_from_name(name, names_dict):
    sig = Sig(names_dict[name])

    # func = sig_to_func(name, names_dict)
    func = sig_to_lambda(sig)

    return func


def sig_to_func(name, names_dict):
    sig = Sig(names_dict[name])

    args = args_str_from_sig(sig)
    # exec(f"global foo_{name}")
    exec(f'global foo_{name}\n\ndef foo_{name}({args}):\n  return None')
    func = eval(f'global foo_{name}')
    return func


def remove_kind(sig, excluded=[VP]):
    params = [param for param in sig.params if param.kind not in excluded]
    return Sig(params)


def dispatch_variadics(sig):
    result = []
    for excluded in [[], [VP], [VK], [VP, VK]]:
        result.append(Sig(_remove_variadics_from_sig(remove_kind(sig, excluded))))

    return result


def sig_with_variadics_to_inputs(sig):
    all_sigs = dispatch_variadics(sig)
    print(all_sigs)
    # for s in all_sigs:
    #     print(list(sig_to_inputs(s)))
    pass



from meshed import FuncNode, DAG
from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep2 import (
    data_from_wav_folder,
    data_from_csv,
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)
from odat.mdat.vacuum import (
    DFLT_ANNOTS_COLS,
    DFLT_CHUNKER,
    DFLT_FEATURIZER,
)
from streamlitfront.examples.util import Graph

from front.spec_maker_base import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from streamlitfront.base import mk_app

DFLT_CHUNKER_MAKER = lambda: DFLT_CHUNKER
DFLT_FEATURIZER_MAKER = lambda: DFLT_FEATURIZER

FixedSizeChunker = DFLT_CHUNKER
Featurizer = DFLT_FEATURIZER


def funcnode_maker(name):
    kwargs = metadata[name]
    return FuncNode(**kwargs)


def dag_maker(funcnames_list):
    func_nodes = list(map(funcnode_maker, funcnames_list))
    return DAG(func_nodes)


metadata = {
    'FixedSizeChunker': {'func': FixedSizeChunker, 'out': 'chks'},
    'FixedSizeChunker100': {'out': 'chks'},
    'ThresholdChunker': {'out': 'chks'},
    'FixedSizeChunkerMaker': {'func': DFLT_CHUNKER_MAKER, 'out': 'chunker'},
    'FeaturizerMaker': {'func': DFLT_FEATURIZER_MAKER, 'out': 'featurizer'},
    'key_fvs_to_tag_fvs': {'func': key_fvs_to_tag_fvs},
    'Featurizer': {'func': DFLT_FEATURIZER, 'out': 'fvs'},
    'store_to_key_fvs': {'func': store_to_key_fvs, 'out': 'key_fvs'},
    'key_fvs_to_tag_fvs': {'func': key_fvs_to_tag_fvs, 'out': 'tag_fv_iterator'},
    'WfStoreMaker': {
        'func': data_from_wav_folder,
        'out': 'wf_store',
        'bind': {'filepath': 'wf_filepath'},
    },
    'AnnotsStoreMaker': {'func': data_from_csv, 'out': 'annots_df'},
    'mk_Xy': {'func': mk_Xy},
}

# Make a dag from only typing info
# my_chunker: Chunker    -->


if __name__ == '__main__':
    # example
    # pyckup
    wf_filepath = '/Users/sylvain/Dropbox/Otosense/VacuumEdgeImpulse/'
    annots_filepath = '/Users/sylvain/Dropbox/sipyb/Testing/data/annots_vacuum.csv'

    funcnames_list = [
        'FixedSizeChunkerMaker',
        'FeaturizerMaker',
        'WfStoreMaker',
        'AnnotsStoreMaker',
        'key_fvs_to_tag_fvs',
        'store_to_key_fvs',
        'key_fvs_to_tag_fvs',
        'mk_Xy',
    ]
    # do a multiselect to do the selection
    # validation, errors to be checked
    dag = dag_maker(funcnames_list)
    print(dag.synopsis_string())

    config_ = {
        APP_KEY: {'title': 'Simple Load and Display'},
        RENDERING_KEY: {DAG: {'graph': {ELEMENT_KEY: Graph, NAME_KEY: 'Flow',},},},
    }

    app = mk_app([dag], config=config_)
    app()



import pandas as pd
import plotly.express as px
import streamlit as st
import numpy as np
from utils import discretize, mk_audio_stream
import time
from taped import LiveWf, simple_chunker
from taped.base import WfChunks
from lined import Pipeline, iterize
from functools import partial

DFLT_CHK_SIZE = 2048 * 4

# simple ML functions
def chk_to_fv(chk):
    return np.std(chk)


def process_chk(chk):
    return chk_to_fv(chk)


featurizer = np.std
chunker = partial(simple_chunker, chk_size=2048)
p = Pipeline(chunker, iterize(featurizer))


# page config
st.set_page_config(
    page_title='Stream a dataframe', layout='wide',
)

# dashboard title
st.title('Stream a dataframe')

# creating a single-element container
placeholder = st.empty()

# live data

i = 0
with LiveWf() as live_wf:
    while True:
        with placeholder.container():
            i += 5
            chks = live_wf[i : i + 100]

            # chks = p(live_wf[i : i + 100])

            dataL = chks
            d = {'sound': dataL}
            # fig = px.line(
            #     data_frame=pd.DataFrame(d),
            #     x=range(200),
            #     y="sound",
            #     range_y=[0, 1000],
            # )
            # st.write(fig)
            st.bar_chart(dataL)



from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any, Iterable, Union
from dol import Files
from meshed import code_to_dag, DAG
from odat.mdat.vacuum import (
    DFLT_ANNOTS_COLS,
    DFLT_CHUNKER,
    DFLT_FEATURIZER,
)
from creek.tools import apply_func_to_index, fanout_and_flatten
import os
import numpy as np
from dol.appendable import appendable
import matplotlib.pyplot as plt

from streamlitfront import mk_app, binder as b

import streamlit as st

import soundfile as sf
import pandas as pd
import zipfile
from io import BytesIO
from pathlib import Path

from dol.appendable import add_append_functionality_to_store_cls
from dol import Store
from dol import FilesOfZip, wrap_kvs, filt_iter

from hear import WavLocalFileStore


DFLT_WF_PATH = '/Users/sylvain/Dropbox/Otosense/VacuumEdgeImpulse/'
DFLT_ANNOT_PATH = '/Users/sylvain/Dropbox/sipyb/Testing/data/annots_vacuum.csv'


def my_obj_of_data(b):
    return sf.read(BytesIO(b), dtype='float32')[0]


@wrap_kvs(obj_of_data=my_obj_of_data)
@filt_iter(filt=lambda x: not x.startswith('__MACOSX') and x.endswith('.wav'))
class WfZipStore(FilesOfZip):
    """Waveform access. Keys are .wav filenames and values are numpy arrays of int16 waveform."""

    pass


def key_to_ext(k):
    _, ext = os.path.splitext(k)
    if ext.startswith('.'):
        ext = ext[1:]
    return ext


def processor_from_ext(ext):
    if ext.startswith('.'):
        ext = ext[1:]
    if ext in {'zip'}:
        pass
    elif ext in {'wav'}:
        pass


def is_zip_file(filepath):
    return zipfile.is_zipfile(filepath)


def is_dir(filepath):
    return os.path.isdir(filepath)


def key_maker(name, prefix):
    return f'{prefix}_{name}'


def wf_store_factory(filepath):
    return store_factory(filepath, data_reader=data_from_wav_folder, prefix='wf_store')


def annot_store_factory(filepath):
    return store_factory(filepath, data_reader=data_from_csv, prefix='annot_store')


def data_from_wav_folder(filepath):
    if is_dir(filepath):
        data = WavLocalFileStore(filepath)

    elif is_zip_file(filepath):
        data = WfZipStore(filepath)

    else:
        raise ('Data not supported')
    return data


def data_from_csv(filepath):
    return pd.read_csv(filepath)


def annot_store_factory(filepath):
    key = key_maker(name=filepath, prefix='annot_store')
    tag = 'annot_store'


def store_factory(filepath, data_reader=data_from_csv, prefix='annot_store'):
    key = key_maker(name=filepath, prefix=prefix)
    tag = prefix

    data = data_reader(filepath)

    return mk_store_item(key, tag, data)


def mk_store_item(key, tag, data):
    return dict(key=key, tag=tag, data=data)


def store_to_key_fvs(wf_store, chunker=DFLT_CHUNKER, featurizer=DFLT_FEATURIZER):
    wf_items = wf_store.items()
    key_chk_tuples = fanout_and_flatten(wf_items, chunker, 1)
    featurizer_iter = partial(apply_func_to_index, apply_to_idx=1, func=featurizer)
    yield from map(featurizer_iter, key_chk_tuples)


def key_fvs_to_tag_fvs(key_fvs, annots_df):
    func = partial(key_to_tag_from_annots, annots_df=annots_df)
    tagger = partial(apply_func_to_index, apply_to_idx=0, func=func)

    yield from map(tagger, key_fvs)


def key_to_tag_from_annots(key, annots_df):
    tag = annots_df['tag'][annots_df['key'] == key].values[0]
    return tag


def mk_Xy(tag_fv_iterator):
    y, X = zip(*list(tag_fv_iterator))
    return np.array(X), y


if __name__ == '__main__':
    pass



from dataclasses import dataclass
from typing import Iterable
from meshed import code_to_dag, DAG
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY
from collections.abc import Callable
from front.crude import prepare_for_crude_dispatch
from streamlitfront.elements import TextInput, SelectBox, FloatSliderInput
from front.elements import OutputBase
from streamlitfront.base import mk_app
from streamlitfront.examples.util import Graph
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
import streamlit as st
import matplotlib.pyplot as plt
import soundfile as sf
from io import BytesIO
from front.crude import Crudifier


param_to_mall_maps = dict(train_audio='train_audio', tag='tag_store')

if 'mall' not in st.session_state:
    st.session_state['mall'] = dict(
        # train_audio={},
        # tag={},
        # unused_store={"to": "illustrate"},
        tag_sound_output={}
    )

mall = st.session_state['mall']
# mall = dict(
#     # train_audio={},
#     # tag={},
#     # unused_store={"to": "illustrate"},
#     tag_sound_output={}
# )


WaveForm = Iterable[int]


@dataclass
class TaggedAudioPlayer(OutputBase):
    def render(self):
        sound, tag = self.output
        if not isinstance(sound, str):
            sound = sound.getvalue()

        st.audio(sound)


@dataclass
class TaggedAudioVisualizer(OutputBase):
    def render(self):
        sound, tag = self.output
        if not isinstance(sound, str):
            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        fig, ax = plt.subplots(figsize=(15, 5))
        ax.plot(arr, label=f'Tag={tag}')
        ax.legend()
        st.pyplot(fig)
        # st.write(arr[:10])


@dataclass
class DfVisualizer(OutputBase):
    def render(self):
        # m = self.output

        st.dataframe(mall['tag_sound_output'])


# crudify = Crudifier(mall=mall, output_store="tag_sound_output")
# f = crudify(display_df)


# May be partialize the mall argument?
# make a separate crudifier, with some defaults specified
# check "verbose"


@Crudifier(mall=mall, output_store='tag_sound_output')
def tag_sound(train_audio: WaveForm, tag: str):
    # mall["tag_store"] = tag
    return (train_audio, tag)


@Crudifier(mall=mall, param_to_mall_map={'result': 'tag_sound_output'})
def display_tag_sound(result):
    return result


@Crudifier(mall=mall, param_to_mall_map={'result': 'tag_sound_output'})
def visualize_tag_sound(result):
    return result


def mall_to_df(mall):
    names = list(mall['tag_sound_output'].keys())
    # pd.DataFrame


def explore_dataset(result):
    return mall['tag_sound_output']


def set_train_test_proportion(train_proportion):
    mall['train_test_proportion'] = train_proportion
    return train_proportion


config_ = {
    APP_KEY: {'title': 'Simple Real Audio ML'},
    # OBJ_KEY: {"trans": crudify},
    RENDERING_KEY: {
        'tag_sound': {
            # "description": {"content": "A very simple learn model example."},
            'execution': {
                'inputs': {
                    'train_audio': {
                        ELEMENT_KEY: MultiSourceInput,
                        'From a file': {ELEMENT_KEY: FileUploader, 'type': 'wav',},
                        'From the microphone': {ELEMENT_KEY: AudioRecorder},
                    },
                    # "tag": {
                    #     ELEMENT_KEY: TextInput,
                    # },
                },
            }
        },
        'display_tag_sound': {
            'execution': {
                'inputs': {
                    'result': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['tag_sound_output'],
                    },
                },
                'output': {ELEMENT_KEY: TaggedAudioPlayer,},
            },
        },
        'visualize_tag_sound': {
            'execution': {
                'inputs': {
                    'result': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['tag_sound_output'],
                    },
                },
                'output': {ELEMENT_KEY: TaggedAudioVisualizer,},
            },
        },
        'explore_dataset': {
            'execution': {
                'inputs': {'result': {ELEMENT_KEY: SelectBox, 'options': 'list',},},
                'output': {ELEMENT_KEY: DfVisualizer,},
            },
        },
        'set_train_test_proportion': {
            'execution': {
                'inputs': {'train_proportion': {ELEMENT_KEY: FloatSliderInput,},}
            },
        },
        DAG: {'graph': {ELEMENT_KEY: Graph, NAME_KEY: 'Flow',},},
        Callable: {
            'execution': {
                'inputs': {
                    'save_name': {ELEMENT_KEY: TextInput, NAME_KEY: 'Save output as',},
                }
            },
        },
    },
}

app = mk_app(
    [tag_sound, display_tag_sound, visualize_tag_sound, set_train_test_proportion],
    config=config_,
)
app()
# st.audio(mall["tag_sound_output"]["s3"][0])
#'execution': {'inputs': {'p': {ELEMENT_KEY: FloatSliderInput,}},}
st.write(mall)



from collections import deque
from itertools import islice
from functools import partial

from lined import Pipe  # pip install lined


def sliding_window(iterable, chk_size):
    # sliding_window('ABCDEFG', 4) -> ABCD BCDE CDEF DEFG
    it = iter(iterable)
    window = deque(islice(it, chk_size), maxlen=chk_size)
    if len(window) == chk_size:
        yield tuple(window)
    for x in it:
        window.append(x)
        yield tuple(window)


def streams_to_named_streams(*streams):
    """
    >>> stream_0 = [1, 2, 3, 4, 5, 6]
    >>> stream_1 = [-1, -2, -3, 0, -5, -6]
    >>> streams_to_named_streams(stream_0, stream_1)
    {0: [1, 2, 3, 4, 5, 6], 1: [-1, -2, -3, 0, -5, -6]}
    """
    return dict(enumerate(streams))


def interleaved_sliding_window_hunks(named_streams, chk_size):
    """
    >>> from functools import partial
    >>> hunker = partial(interleaved_sliding_window_hunks, chk_size=2)
    >>> named_streams = {0: [1, 2, 3, 4, 5, 6], 1: [-1, -2, -3, 0, -5, -6]}
    >>> assert list(hunker(named_streams)) == [
    ...     ((1, -1), (2, -2)),
    ...     ((2, -2), (3, -3)),
    ...     ((3, -3), (4, 0)),
    ...     ((4, 0), (5, -5)),
    ...     ((5, -5), (6, -6))
    ... ]


    """
    return sliding_window(zip(*named_streams.values()), chk_size)


def indexed_mapper(slabs, slab_func, indexer=enumerate):
    """
    >>> indexed_apply = partial(
    ...     indexed_mapper, slab_func=lambda x: x * 10, indexer=enumerate)
    >>> list(indexed_apply([-1, -2, -3, -4]))
    [(0, -10), (1, -20), (2, -30), (3, -40)]
    """
    for i, slab in indexer(slabs):
        func_result = slab_func(slab)
        yield i, func_result


stream_0 = [1, 2, 3, 4, 5, 6]
stream_1 = [-1, -2, -3, 0, -5, -6]
# list(event_detection_filter(click_channel_0, click_channel_1, min_lag=1, max_lag=1))


def mk_pipe(
    *, chk_size=1, slab_func=lambda x: x, filt=lambda x: True, indexer=enumerate
):
    return Pipe(
        streams_to_named_streams,
        hunker=partial(interleaved_sliding_window_hunks, chk_size=chk_size),
        indexed_tags=partial(indexed_mapper, slab_func=slab_func, indexer=indexer),
        user_filter=partial(filter, filt),
    )


def test_mk_pipe():
    p = mk_pipe(chk_size=2, slab_func=lambda x: any(map(sum, x)), indexer=enumerate)
    assert list(p(stream_0, stream_1)) == [
        (0, False),
        (1, False),
        (2, True),
        (3, True),
        (4, False),
    ]

    p = mk_pipe(
        chk_size=2, slab_func=lambda x: any(map(sum, x)), filt=lambda x: x[1] is True
    )

    assert list(p(stream_0, stream_1)) == [(2, True), (3, True)]


# test_mk_pipe()



"""A simple app to demo the use of Mappings to handle complex type"""
from typing import Mapping

ComplexType = float  # just pretend it's complex!


def func(salary: ComplexType, n_months: int = 12):
    return salary * n_months


SalaryKey = str  # or some type that will resolve in store-fed key selector
SalaryMapping = Mapping[SalaryKey, ComplexType]

salary_store: SalaryMapping
salary_store = {'sylvain': 10000, 'christian': 2000, 'thor': 50000}


def store_wrapped_func(salary: SalaryKey, n_months: int = 12):
    salary = salary_store[salary]
    return func(salary, n_months)


assert store_wrapped_func('sylvain', 6) == 60000 == func(10000, 6)


if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs

    app = dispatch_funcs([func, store_wrapped_func])
    app()



from collections import defaultdict
from functools import wraps, partial, reduce
from io import BytesIO
from itertools import chain
from math import ceil
from operator import itemgetter
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    Callable,
    MutableMapping,
    Protocol,
    Iterator,
    DefaultDict,
    List,
)
from wave import Wave_write

from dol import appendable, Files, wrap_kvs
from i2 import Sig
from mongodol.tracking_methods import (
    track_method_calls,
    TrackableMixin,
)
from recode import (
    encode_pcm_bytes,
    decode_wav_bytes,
)


class BulkStore(Protocol):
    """Mutable Mapping with Append Method"""

    def append(self, item):
        pass

    def __setitem__(self, __k, __v) -> None:
        pass

    def __delitem__(self, __v) -> None:
        pass

    def __getitem__(self, __k):
        pass

    def __len__(self) -> int:
        pass

    def __iter__(self) -> Iterator:
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


audio_slab_kv = itemgetter('timestamp', 'wf')


N_TO_BULK = 10


def merge_append_ts_wf(items: List[dict]):
    """Take a list of kwargs for method append(item) and return a single kwargs

    :param items: [{'item': {'timestamp': 0, 'wf': [0, 0]}},
                   {'item': {'timestamp': 1, 'wf': [1, 1]}},
                   ...]
    :return: {'item': {'timestamp': 0, 'wf': [0, 0, 1, 1, ...]}}
    """
    ts = items[0]['item']['timestamp']
    joined_wf = list(
        chain.from_iterable(
            map(
                partial(
                    reduce, lambda x, y: y(x), (itemgetter('item'), itemgetter('wf')),
                ),
                items,
            )
        )
    )
    return {'item': {'timestamp': ts, 'wf': joined_wf}}


class CountMergeExecute(TrackableMixin):
    """Make tracks factory count method calls when order of different methods is not important.
    Apply a merging method to join all arguments before executing tracked method once.
    """

    tracks_factory: DefaultDict[Callable, List[tuple]] = partial(defaultdict, list)

    _MERGE_KWARGS_MAP = {'append': merge_append_ts_wf}

    def _execute_method_tracks(self, method: Callable):
        if merge_kwargs_method := self._MERGE_KWARGS_MAP.get(method.__name__):
            sig = Sig(method)
            kwargs_list = [
                sig.kwargs_from_args_and_kwargs((self, *a), kw)
                for a, kw in self._tracks[method]
            ]
            merged_kwargs = merge_kwargs_method(kwargs_list)
            method(self, **merged_kwargs)
        else:
            for args, kwargs in self._tracks[method]:
                method(self, *args, **kwargs)

    def _execute_tracks(self):
        for method in self._tracks:
            self._execute_method_tracks(method)


def track_n_calls_of_method_then_execute(method: Callable, n_calls=N_TO_BULK):
    @wraps(method)
    def tracked_method(self: CountMergeExecute, *args, **kwargs):
        self._tracks[method].append((args, kwargs))
        if len(self._tracks[method]) >= n_calls:
            self._execute_method_tracks(method)
            del self._tracks[method]

    return tracked_method


def merge_and_write_upon_count(
    store_cls: MutableMapping = None,
    *,
    tracked_methods='append',
    tracking_mixin=CountMergeExecute,
    tracking_merge_map=None,
    calls_tracker=track_n_calls_of_method_then_execute,
    n_calls_to_execute=N_TO_BULK,
    item2kv=audio_slab_kv,
    key_of_id=None,
    id_of_key=None,
    obj_of_data=None,
    data_of_obj=None,
):
    """Equivalent of using the following 3 decorators

    @track_method_calls(
        tracked_methods='append',
        tracking_mixin=CountMergeExecute,
        calls_tracker=track_n_calls_of_method_then_execute,
    )
    @appendable(item2kv=audio_slab_kv)
    @wrap_kvs(
        key_of_id=filename_to_ts,
        id_of_key=ts_to_filename,
        obj_of_data=wav_to_wf,
        data_of_obj=wf_to_wav,
    )
    class WavFileStore(Files, BulkStore):
        pass

    :param store_cls:
    :param tracked_methods:
    :param tracking_mixin:
    :param tracking_merge_map:
    :param calls_tracker:
    :param n_calls_to_execute:
    :param item2kv:
    :param key_of_id:
    :param id_of_key:
    :param obj_of_data:
    :param data_of_obj:
    :return:
    """

    def _decorator(cls):
        if isinstance(tracking_merge_map, MutableMapping):
            _tracking_mixin = type(
                tracking_mixin.__name__,
                (tracking_mixin,),
                {'_MERGE_KWARGS_MAP': tracking_merge_map},
            )
        else:
            _tracking_mixin = tracking_mixin

        return track_method_calls(
            appendable(
                wrap_kvs(
                    cls,
                    key_of_id=key_of_id,
                    id_of_key=id_of_key,
                    obj_of_data=obj_of_data,
                    data_of_obj=data_of_obj,
                ),
                item2kv=item2kv,
            ),
            tracked_methods=tracked_methods,
            tracking_mixin=_tracking_mixin,
            calls_tracker=partial(calls_tracker, n_calls=n_calls_to_execute),
        )

    return _decorator(store_cls) if callable(store_cls) else _decorator


def ts_to_filename(ts, ext='.wav'):
    return f'{ts}{ext}'


def filename_to_ts(name, ext='.wav'):
    return int(name[: -len(ext)])


def wf_to_wav(
    wf, sr: int = 44100, width_bytes: int = 2, *, n_channels: int = 1
) -> bytes:
    bio = BytesIO()
    with Wave_write(bio) as obj:
        obj.setparams((n_channels, width_bytes, sr, 0, 'NONE', 'not compressed'))
        wf_bytes = encode_pcm_bytes(wf, width=width_bytes * 8, n_channels=n_channels)
        obj.writeframes(wf_bytes)
        bio.seek(0)
    return bio.read()


def wav_to_wf(wav):
    wf, sr = decode_wav_bytes(wav)
    return wf


@merge_and_write_upon_count(
    key_of_id=filename_to_ts,
    id_of_key=ts_to_filename,
    obj_of_data=wav_to_wf,
    data_of_obj=wf_to_wav,
)
class WavFileStore(Files, BulkStore):
    pass


def _test_merge_append_ts_wf():
    a = []
    for i in range(10):
        a.append({'item': {'timestamp': i, 'wf': [i] * 2}})
    ma = merge_append_ts_wf(a)
    assert 'item' in ma
    assert 'timestamp' in ma['item']
    assert 'wf' in ma['item']


def _test_dict_store():
    """Test decorator usage with basic dict store mapping"""

    @merge_and_write_upon_count
    class DictStore(dict, BulkStore):
        pass

    _test_store(DictStore())

    @merge_and_write_upon_count(n_calls_to_execute=7)
    class DictStore7(dict, BulkStore):
        pass

    _test_store(DictStore7(), n_bulk=7)


def _test_files_store():
    """Test decorator usage with file store"""

    with TemporaryDirectory() as tmpdirname:
        audio_store = WavFileStore(rootdir=tmpdirname)
        _test_store(audio_store)
        print(f'{len(list(Path(tmpdirname).iterdir()))=} == {len(audio_store)=}')
        assert len(list(Path(tmpdirname).iterdir())) == len(audio_store)


def _test_store(
    store_instance: MutableMapping, *, n=22, chk_size=2, log=print, n_bulk=N_TO_BULK
):
    """Test append, tracking, input, and output usage

    :param store_instance:
    :param n: number of chunks
    :param chk_size: samples per chunk
    :param log: print
    :return:
    """

    def n_append_tracks(store):
        _, append_track = next(iter(store._tracks.items()), (None, tuple()))
        return len(append_track)

    with store_instance as a:
        for i in range(n):
            a.append({'timestamp': i, 'wf': [i] * chk_size})

            _, append_track = next(iter(a._tracks.items()), (None, tuple()))

            log('print#2', f'{i=}', f'{n_append_tracks(a)=}', a)
            assert (
                n_append_tracks(a) == (i + 1) % n_bulk
            ), 'Tracking should auto flush when size limit is reached'
            log('print#3', f'{len(a) == 0=}')
            assert len(a) == (i + 1) // n_bulk, list(a)
            log(a._tracks)

        _, append_track = next(iter(a._tracks.items()), (None, tuple()))

        log('print#4', f'{n_append_tracks(a) == n=}')
        assert n_append_tracks(a) == n % n_bulk
        log('print#5', a, len(append_track))

    _, append_track = next(iter(a._tracks.items()), (None, tuple()))

    log('print#6', n_append_tracks(a), a)
    assert n_append_tracks(a) == 0
    assert len(a) == ceil(n / n_bulk), f'{len(a)=} == {ceil(n / n_bulk)=}'
    for i, (k, v) in enumerate(sorted(a.items())):
        log('print#7', f'{(k, v)=}')
        assert k == i * n_bulk
        if k + n_bulk > n:
            assert (
                len(v) / chk_size == n - k
            ), f'{i=}, {k=}, {len(v) / chk_size=}, {n - k=}'
        else:
            assert (
                len(v) / chk_size == n_bulk
            ), f'{i=}, {k=}, {len(v) / chk_size=}, {n_bulk=}'
        v_iter = iter(v)
        for j in range(i * n_bulk, min((i + 1) * n_bulk, n)):
            for _ in range(chk_size):
                assert (
                    next(v_iter) == j
                ), f'{i=}, {j=}, {k=}, {i * n_bulk=}, {min((i + 1) * n_bulk, n)=}'

    print(f'Test passed with params: {store_instance=}, {n=}, {chk_size=}, {log=}')


if __name__ == '__main__':
    _test_merge_append_ts_wf()
    _test_dict_store()
    _test_files_store()



import websocket
import json
from typing import Any
from collections import deque

# use SourceReader, check the audio one (audiostream2py)


def on_message(self):
    def wrapped(
        ws, message
    ):  # add incoming data to a standard deque: standard attribute to the SourceReader, so the read would read from the deque and pop data
        result = json.loads(message)
        values = result['values']
        bt = result['timestamp']
        x = values[0]
        d = {'bt': bt, 'acc_x': x}
        self.data.append(d)
        # source.emit(x)
        # print(d)

    return wrapped


def on_error(ws, error):
    print('error occurred')
    print(error)


def on_close(ws, close_code, reason):
    print('connection close')
    print('close code : ', close_code)
    print('reason : ', reason)


def on_open(ws):
    print('connection open')


DFLT_URL = 'ws://192.168.1.3:8080/sensor/connect?type=android.sensor.accelerometer'
DFLT_DEQUE_SIZE = 5


class AccelSource:
    """The interface for the StreamSource class."""

    def __init__(self, url, deque_size=DFLT_DEQUE_SIZE):
        self.url = url
        self.data = deque(maxlen=deque_size)

    def key(self, data) -> Any:
        ...

    @property
    def info(self) -> dict:
        return {}

    @property
    def sleep_time_on_read_none_s(self) -> int:
        return 0

    def read(self) -> Any:
        data = self.data.popleft()
        print(f'data retrieved = {data}')
        return data

    def open(self) -> None:
        self.ws = websocket.WebSocketApp(  # put it in open
            self.url,  # "ws://192.168.1.1:56478/sensor/connect?type=android.sensor.accelerometer",
            on_open=on_open,
            on_message=on_message(self),
            on_error=on_error,
            on_close=on_close,
        )
        self.ws.run_forever()

    def close(self) -> None:
        self.ws.close()


if __name__ == '__main__':
    stream = AccelSource(url=DFLT_URL)
    stream.open()
    stream.read()
    stream.close()



# Usage: streamlit run barfi_app.py

from barfi import st_barfi, Block

feed = Block(name='Feed')
feed.add_input()
result = Block(name='Result')
result.add_output()

barfi_result = st_barfi(base_blocks=[feed, result])



import functools
import random
import time

import pandas as pd

from streamz import Stream
import hvplot.streamz
from river_t import RiverTrain
from river import cluster
import holoviews as hv
from panel.pane.holoviews import HoloViews
import panel as pn

hv.extension('bokeh')

model = cluster.KMeans(n_clusters=3, sigma=0.1, mu=0.5)
centres = [[random.random(), random.random()] for _ in range(3)]
count = [0]


def gen(move_chance=0.05):
    centre = int(random.random() * 3)  # 3x faster than random.randint(0, 2)
    if random.random() < move_chance:
        centres[centre][0] += random.random() / 5 - 0.1
        centres[centre][1] += random.random() / 5 - 0.1
    value = {
        'x': random.random() / 20 + centres[centre][0],
        'y': random.random() / 20 + centres[centre][1],
    }
    count[0] += 1
    return value


def get_clusters(model):
    # return [{"x": xcen, "y": ycen}, ...] for each centre
    data = [{'x': v['x'], 'y': v['y']} for k, v in model.centers.items()]
    return pd.DataFrame(data, index=range(3))


def main(viz=True):
    cadance = 0.01

    ex = pd.DataFrame({'x': [0.5], 'y': [0.5]})
    pipe_in = hv.streams.Pipe(data=ex)
    pipe_out = hv.streams.Pipe(data=ex)

    # setup pipes
    s = Stream.from_periodic(gen, cadance)

    # Branch 0: Input/Observations
    obs = s.map(lambda x: pd.DataFrame([x]))

    # Branch 1: Output/River ML clusters
    km = RiverTrain(model, pass_model=True)
    s.map(lambda x: (x,)).connect(km)  # learn takes a tuple of (x,[ y[, w]])
    clusters = km.map(get_clusters)

    concat = functools.partial(pd.concat, ignore_index=True)

    def accumulate(previous, new, last_lines=50):
        return concat([previous, new]).iloc[-last_lines:, :]

    partition_obs = 10
    particion_clusters = 10
    backlog_obs = 100

    # .partition is used to gather x number of points
    # before sending them to the plots
    # .accumulate allows to generate a backlog

    (
        obs.partition(partition_obs)
        .map(concat)
        .accumulate(functools.partial(accumulate, last_lines=backlog_obs))
        .sink(pipe_in.send)
    )
    (clusters.partition(particion_clusters).map(pd.concat).sink(pipe_out.send))

    # start things
    s.emit(gen())  # set initial model
    for i, (x, y) in enumerate(centres):
        model.centers[i]['x'] = x
        model.centers[i]['y'] = y

    print('starting')

    if viz:
        # plot
        button_start = pn.widgets.Button(name='Start')
        button_stop = pn.widgets.Button(name='Stop')

        t0 = 0

        def start(event):
            s.start()
            global t0
            t0 = time.time()

        def stop(event):
            print(count, 'events')
            global t0
            t_spent = time.time() - t0
            print('frequency', count[0] / t_spent, 'Hz')
            print('Current centres', centres)
            print('Output centres', [list(c.values()) for c in model.centers.values()])
            s.stop()

        button_start.on_click(start)
        button_stop.on_click(stop)

        scatter_dmap_input = hv.DynamicMap(hv.Scatter, streams=[pipe_in]).opts(
            color='blue'
        )
        scatter_dmap_output = hv.DynamicMap(hv.Scatter, streams=[pipe_out]).opts(
            color='red'
        )
        pl = scatter_dmap_input * scatter_dmap_output
        pl.opts(xlim=(-0.2, 1.2), ylim=(-0.2, 1.2), height=600, width=600)

        pan = HoloViews(pl)
        app = pn.Row(pn.Column(button_start, button_stop), pan)
        app.show()
    else:
        s.start()
        time.sleep(5)
        print(count, 'events')
        print('frequency', count[0] / 5, 'Hz')
        print('Current centres', centres)
        print('Output centres', [list(c.values()) for c in model.centers.values()])
        s.stop()


if __name__ == '__main__':
    main(viz=True)



"""A simple app to demo the use of Mappings to handle complex type"""
from typing import Mapping
from enum import Enum


ComplexType = float  # just pretend it's complex!


def func(salary: ComplexType, n_months: int = 12) -> float:
    return salary * n_months


SalaryKey = str  # or some type that will resolve in store-fed key selector
SalaryMapping = Mapping[SalaryKey, ComplexType]

salary_store: SalaryMapping
salary_store = {'sylvain': 10000, 'christian': 2000, 'thor': 50000}


def mk_choices_from_store(store):
    choices = Enum('Choices', {key: key for key in store.keys()})
    return choices


if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from pydantic import BaseModel, Field, ValidationError, parse_obj_as
    from streamlitfront.page_funcs import SimplePageFuncPydanticWrite

    class ChoiceModel(BaseModel):
        single_selection: mk_choices_from_store(salary_store) = Field(
            ..., description='Only select a single item from a set.'
        )

    configs = {'page_factory': SimplePageFuncPydanticWrite}

    def wrapped_func(selection: ChoiceModel, n_months: int):
        # print(f"{type(selection)=}, {selection=}")
        selection = selection.single_selection.value
        salary_val = salary_store[selection]

        return func(salary_val, n_months)

    app = dispatch_funcs([func, wrapped_func], configs=configs)
    app()



import streamlit as st
from hear import WavLocalFileStore
from plunk.sb.front_experiments.data_visualizer.utils.tools import (
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)
import umap
import umap.plot
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


root_dir = '/Users/sylvain/Dropbox/_odata/sound/vacuum'
wf_store = WavLocalFileStore(root_dir)
annots_path = '../data/annots_vacuum.csv'
df = pd.read_csv(annots_path)


key_fvs = store_to_key_fvs(wf_store)
tag_fv_iterator = key_fvs_to_tag_fvs(key_fvs, annots_df=df)
X, y = mk_Xy(tag_fv_iterator)

st.title('visualizer')


def plot_umap(X, y, show_legend=True):
    mapper = umap.UMAP().fit(X)
    fig, ax = plt.subplots()

    umap.plot.points(mapper, labels=np.array(y), show_legend=show_legend, ax=ax)
    st.pyplot(fig)


plot_umap(X, y)



import io
import numpy as np
import soundfile as sf
import streamlit as st
import pyaudio


def discretize(arr, num_windows=200, func=np.mean):
    step = len(arr) // num_windows
    dis = [func(arr[i * step : i * (step + 1)]) for i in range(num_windows)]
    return np.array(dis)


def wav_complex_display(uploaded_file):
    bytes = bytes_from_uploaded(uploaded_file)
    st.audio(bytes)
    arr = arr_from_bytes(bytes)
    arr_d = discretize(arr)
    st.write(f'length = {len(arr)}')
    st.bar_chart(arr_d)


def arr_from_bytes(bytes, dtype='int16'):
    arr = sf.read(bytes, dtype=dtype)[0]

    return arr


def bytes_from_uploaded(uploaded_file):
    uploaded_file.seek(0)
    result = io.BytesIO(uploaded_file.read())

    return result


def mk_audio_stream():
    p = pyaudio.PyAudio()
    stream = p.open(
        format=pyaudio.paInt16,
        channels=1,
        rate=44100,
        input=True,
        output=True,
        frames_per_buffer=1024,
        output_device_index=3,
    )
    return stream



from IPython.core.magic import (
    Magics,
    magics_class,
    line_magic,
    cell_magic,
    line_cell_magic,
)

import ast


def print_ast_or_ignore(src, src_name='cell'):
    try:
        print(ast.parse(src))
    except Exception:
        print(f"Couldn't parse this {src_name}")
        pass


# The class MUST call this class decorator at creation time
@magics_class
class MyMagics(Magics):
    @line_cell_magic
    def dispdag(self, line, cell=None):
        'Magic that works both as %lcmagic and as %%lcmagic'
        from meshed import code_to_dag

        src = cell or line  # take line if cell is None
        dag = code_to_dag(src)
        self.shell.user_ns[dag.__name__] = dag
        return dag.dot_digraph()

    @line_magic
    def lmagic(self, line):
        'my line magic'
        print('Full access to the main IPython object:', self.shell)
        print('')
        print(f'{dir(self.shell)=}')
        print('')
        print('Variables in the user namespace:', list(self.shell.user_ns.keys()))
        return line

    @cell_magic
    def cmagic(self, line, cell):
        'my cell magic'
        return line, cell

    @line_cell_magic
    def lcmagic(self, line, cell=None):
        'Magic that works both as %lcmagic and as %%lcmagic'
        if cell is None:
            # print("Called as line magic")
            print_ast_or_ignore(line, src_name='line')
            return line
        else:
            print('Called as cell magic')
            print_ast_or_ignore(cell, src_name='cell')
            return line, cell


# In order to actually use these magics, you must register them with a
# running IPython.


def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ipython.register_magics(MyMagics)



import pytest
from plunk.sb.sig_compatibility.sig_code_to_dag import is_compatible
from i2 import Sig


def test_dummy():
    return True


def f(w, ww, www):
    return w * ww + www


def f_new(www, /, ww=2, *, w: int = 1):
    return w * ww + www


def test_comp_sig():
    assert is_compatible(f, Sig(f_new))



from i2 import Sig


def matcher_by_index(sig1, sig2: Sig):
    from itertools import zip_longest

    result = list(zip_longest(sig1.params, sig2.params, fillvalue=None))
    min_length = min(len(sig1), len(sig2))

    return result[:min_length], result[min_length:]


if __name__ == '__main__':
    sig1 = Sig('x y z')
    sig2 = Sig('a b')
#     result = matcher_by_index(sig1, sig2)
#     expected = ([(<Param "x">, <Param "a">), (<Param "y">, <Param "b">)],
#  [(<Param "z">, None)])
#     assert result == expected



import asyncio
from websockets import connect
from stream2py.simply import mk_stream_buffer
import pandas as pd
import streamlit as st
import json
from collections import deque

df = pd.DataFrame()
df['value'] = 0.0

URI = 'ws://192.168.1.3:8081/sensor/connect?type=android.sensor.accelerometer'


def init_deque(maxlen=20):
    zeros = [0 for _ in range(maxlen)]
    d = deque(zeros, maxlen=maxlen)
    return d


st.title('Accelerometer')

data_deque = init_deque()
df['value'] = data_deque
# st.session_state["d"] = df
# st.write(st.session_state["d"])


async def accelerometer(uri=URI):
    # st.write(df)
    async with connect(uri) as websocket:
        while True:
            data = await websocket.recv()
            result = json.loads(data)['values'][0]
            data_deque.append(result)
            df['value'] = data_deque
            # st.session_state["d"] = df
            # st.write(data_deque)
            # st.write(df)

            # print(data)
            # with mk_stream_buffer(
            #     read_stream=lambda open_inst: data,
            #     open_stream=lambda: print("open"),
            #     close_stream=lambda open_inst: print("close"),
            #     auto_drop=False,
            #     maxlen=200,
            # ) as count_stream_buffer:
            #     count_reader = count_stream_buffer.mk_reader()
            #     print(f"start reading {count_reader.source_reader_info}")
            #     for i in count_reader:
            #         # do stuff with read data
            #         print(i)
            #         if count_stream_buffer.auto_drop is False:
            #             count_stream_buffer.drop()

            #     print("done reading")


# test = st.empty()
# status = st.empty()
# start = st.checkbox("Connect to WS Server")

# asyncio.run(accelerometer(test))
start = True
if start:
    asyncio.run(accelerometer())

placeholder = st.empty()
with placeholder.container():
    st.write(df)



# ---------------------------------------------------------------------------------------
# The following is to demo that (in a notebook at least) when we import this module,
# the decorated functions don't appear in tab-completion list, though they ARe
# attributes of the module object.


# This one is tab-complete visible
def bar(x):
    return x


# 1 ----------
def my_deco(func):
    print(f'{func.__name__} is being called')

    def wrapped_func(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapped_func


# Not even importable!!!
@my_deco
def foo_my_deco(x):
    return x


# 2 ----------
from functools import wraps


def foo(x):  # this will be overwritten by foo_wraps which will be given the foo name
    return x


# Not tab-complete visible
@wraps(foo)
def foo_wraps(x):
    return x


# 3 ----------
from i2.wrapper import wrap


# Not tab-complete visible
@wrap
def foo_wrap(x):
    return x


# 4 ----------
from i2.deco import double_up_as_factory


# Not tab-complete visible
@double_up_as_factory
def foo_double_up_as_factory(x=None):
    return x


# test code (to be run in notebook
def test_decorated_funcs():
    """Meant to be copied"""
    # this shows that they exist (but)

    import plunk.tw.things_decorators_break as m

    expected_func_names = (
        'foo_my_deco',
        'foo_wraps',
        'foo_wrap',
        'foo_double_up_as_factory',
    )
    assert all(x in dir(m) for x in expected_func_names)
    dir(m)

    # and we can import them with from... import... idiom:
    from plunk.tw.things_decorators_break import (
        foo_my_deco,
        foo_wraps,
        foo_wrap,
        foo_double_up_as_factory,
    )

    # But in jupyter notebook, can't tab-complete with that idiom (but can tab
    # complete bar)



"""Tool to get a timestamped transcription of a video file.

Supports .mov files.


"""

import moviepy.editor as mp
import speech_recognition as sr


def mov_to_transcription(mov_filepath: str):
    """
    Transcribes the audio from a MOV file using Google Speech Recognition API.

    :param mov_filepath: Filepath of .mov file
    :return: A dict whose keys are timestamps (seconds offset) and values are
    transcribed text
    """
    # Note that this implementation uses a temporary WAV file to store the audio
    # extracted from the MOV file. This is necessary because the SpeechRecognition
    # library only supports audio files in certain formats, including WAV.
    # Once the audio has been transcribed, the temporary file is deleted.
    # TODO: Use an actual temp file (file not even deleted here).

    # Load the MOV file and extract the audio
    video = mp.VideoFileClip(mov_filepath)
    audio = video.audio.to_audiofile('temp.wav')

    # Transcribe the audio using Google Speech Recognition API
    recognizer = sr.Recognizer()
    with sr.AudioFile('temp.wav') as source:
        audio = recognizer.record(source)
    transcription = recognizer.recognize_google(audio)

    # Convert the transcription to a dictionary of timestamped text
    words = transcription.split()
    duration = video.duration
    # TODO: Perhaps int (floor) better than round?
    timestamps = [round((i / len(words)) * duration, 2) for i in range(len(words))]
    result = {timestamps[i]: words[i] for i in range(len(words))}

    return result


def mov_to_mp4(
    mov_filepath: str,
    output_file_path=None,
    output_format='mp4',
    output_codec='libx264',
    audio_codec='acc',
):
    from moviepy.editor import VideoFileClip

    if output_file_path is None:
        import os

        output_file_path = os.path.splitext(mov_filepath)[0] + f'.{output_format}'

    # Load the video file
    video = VideoFileClip(mov_filepath)

    # Convert the video to mp4 format
    video.write_videofile(
        output_file_path, codec=output_codec, audio=True, audio_codec=audio_codec
    )

    # Close the video file
    video.close()



from dataclasses import dataclass
from functools import partial
import re
from typing import Iterable, List
from streamlitfront import mk_app, binder as b
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from front.crude import Crudifier
from streamlitfront.elements import FileUploader, SuccessNotification
import re
import uuid

from plunk.ap.snippets import get_mall

SEARCH_MODE_KEYS = 1
SEARCH_MODE_VALUES = 2
SEARCH_MODE_KEYS_AND_VALUES = 3
SEARCH_MODE_PATH = 4

DOUBLE_WILDCARD = str(uuid.uuid4())


_mall = dict(
    search_mode={
        'Keys Only': SEARCH_MODE_KEYS,
        'Values Only': SEARCH_MODE_VALUES,
        'Keys and Values': SEARCH_MODE_KEYS_AND_VALUES,
        'Path': SEARCH_MODE_PATH,
    },
    sound_output=dict(),
    data=dict(
        aaa=1,
        bbb='2',
        ccc=[3],
        ddd={'4': 4, '5': [{'e': 6}, {'f': 7}, ['g', 'h', 'i']]},
        jjj=[{'k': 8, 'l': 9}, 'm', 10],
    ),
)
mall = get_mall(_mall)
crudifier = partial(Crudifier, mall=mall)

WaveForm = Iterable[int]


@crudifier(output_store='sound_output')
def upload_sound(train_audio: List[WaveForm], tag: str):
    return train_audio, tag


@dataclass
class DictExplorer:
    dict_to_explore: dict

    def __call__(
        self,
        search_mode: int,
        search: str,
        match_case: bool = False,
        match_whole_word: bool = False,
        use_regex: bool = False,
    ):
        if not search:
            return self.dict_to_explore

        self.match_case = match_case
        self.match_whole_word = match_whole_word
        self.use_regex = use_regex

        self.search = search if self.match_case else search.lower()

        if not self.search:
            return self.dict_to_explore

        if not self.match_case:
            self.search = search.lower()

        self.search_keys = search_mode in (
            SEARCH_MODE_KEYS,
            SEARCH_MODE_KEYS_AND_VALUES,
            SEARCH_MODE_PATH,
        )
        self.search_values = search_mode in (
            SEARCH_MODE_VALUES,
            SEARCH_MODE_KEYS_AND_VALUES,
        )

        return (
            self._search_path()
            if search_mode == SEARCH_MODE_PATH
            else self._filter_dict_from_str()
        )

    def _search_path(self):
        self.match_whole_word = True
        self.use_regex = True
        self.path = [
            re.sub(  # add a dot before any wildcard with no dot to use it as a regex ('*' -> '.*')
                r'(?<!\.)\*',
                '.*',
                re.sub(  # escape the double wildcards first ('**' -> value of DOUBLE_WILDCARD)
                    r'^\*\*$', DOUBLE_WILDCARD, el
                ),
            )
            for el in re.split(
                r'(?<!\\)\/', self.search
            )  # (?<!\\) is for escaping slash characters using a backslash ('\/')
            if el
        ]
        return self._filter_dict_from_path()

    def _filter_dict_from_path(self, d=None, path=None, double_wildcard=False):
        def build_dict():
            _double_wildcard = double_wildcard
            for k, v in d.items():
                _path_tail = path_tail
                match = False
                if next_el == DOUBLE_WILDCARD:
                    _double_wildcard = True
                else:
                    match = self._include_item(next_el, k)
                    if match:
                        _double_wildcard = False
                    else:
                        _path_tail = [next_el] + path_tail
                if match or _double_wildcard:
                    if isinstance(v, dict):
                        _v = self._filter_dict_from_path(
                            v, _path_tail, _double_wildcard
                        )
                        if _v != {} or (v == {} and not _path_tail):
                            yield k, _v
                    elif not _path_tail:
                        yield k, v

        d = self.dict_to_explore if d is None else d
        path = self.path if path is None else path
        if not path:
            return d
        next_el = path[0]
        path_tail = path[1:]
        return {k: v for k, v in build_dict()}

    def _filter_dict_from_str(self, d=None):
        def build_dict():
            for k, v in d.items():
                if isinstance(v, dict):
                    if self._include_item(self.search, k):
                        yield k, v
                    else:
                        _v = self._filter_dict_from_str(v)
                        if _v:
                            yield k, _v
                elif self._include_item(self.search, k, v):
                    yield k, v

        d = self.dict_to_explore if d is None else d
        return {k: v for k, v in build_dict()}

    def _include_item(self, search, key, value=''):
        _key = str(key)
        _value = str(value)
        if not self.match_case:
            _key = _key.lower()
            _value = _value.lower()
        if self.use_regex:
            regex = f'^{search}$' if self.match_whole_word else search
            return (self.search_keys and re.search(regex, _key)) or (
                self.search_values and re.search(regex, _value)
            )
        else:
            if self.match_whole_word:
                return (self.search_keys and search == _key) or (
                    self.search_values and search == _value
                )
            return (self.search_keys and search in _key) or (
                self.search_values and search in _value
            )


@crudifier(param_to_mall_map=['search_mode'])
def explore_mall(
    search_mode: int,
    search: str,
    match_case: bool = False,
    match_whole_word: bool = False,
    use_regex: bool = False,
):
    dict_explorer = DictExplorer(mall)
    return dict_explorer(
        search_mode=search_mode,
        search=search,
        match_case=match_case,
        match_whole_word=match_whole_word,
        use_regex=use_regex,
    )


# def on_select_search_mode():
#     print(b.selected_search_mode())
#     b.match_whole_word.set(b.selected_search_mode() != 4)


if __name__ == '__main__':

    app = mk_app(
        [upload_sound, explore_mall],
        config={
            APP_KEY: {'title': 'Store Explorer'},
            RENDERING_KEY: {
                'upload_sound': {
                    NAME_KEY: 'Data Loader',
                    'execution': {
                        'inputs': {
                            'train_audio': {
                                ELEMENT_KEY: FileUploader,
                                'type': 'wav',
                                'accept_multiple_files': True,
                            },
                        },
                        'output': {
                            ELEMENT_KEY: SuccessNotification,
                            'message': 'Wav files loaded successfully.',
                        },
                    },
                },
                'explore_mall': {
                    'execution': {
                        # 'inputs': {
                        #     'search_mode': {
                        #         'value': b.selected_search_mode,
                        #         # 'on_value_change': on_select_search_mode,
                        #     },
                        #     'match_whole_word': {
                        #         'value': b.match_whole_word,
                        #         'disabled': b.selected_search_mode() == 'Path'
                        #     }
                        # },
                        'auto_submit': True,
                    }
                },
            },
        },
    )
    app()



import numpy as np

from functools import partial
from typing import Iterable, List
from lined import LineParametrized

import soundfile as sf
from io import BytesIO
import streamlit as st

###### ML Tools ###########################
from omodel.gen_utils.chunker import fixed_step_chunker
from omodel.outliers.pystroll import OutlierModel as Stroll

DFLT_CHK_SIZE = 2048
DFLT_CHK_STEP = 2048


def grouper(arr, n):
    """
    Splits inputs into n consecutive non-overlapping groups
    """
    return np.array_split(arr, n)


def apply_func_to_index_groups(func, arr, idx_groups):
    """
    Applies func to the slices of array arr specified by the groups
    """
    result = [func(arr[gp]) for gp in idx_groups]

    return result


DFLT_WINDOW_OUTLIER = 30
DFLT_NUM_OUTLIERS = 3


def arg_top_max(arr, num_elements):
    """
    returns the largest num_elements from the array arr
    """
    result = [arr.index(i) for i in sorted(arr, reverse=True)][:num_elements]
    return result


from more_itertools import consecutive_groups


def indices_for_percentile(arr, low=0, high=100):
    return np.percentile(arr, [low, high])


def consecutive_indices(arr):
    return list(map(list, consecutive_groups(arr)))


def get_groups_extremities_all(
    arr, groups_indices, func=np.mean, num_outliers=DFLT_NUM_OUTLIERS,
):

    means = apply_func_to_index_groups(func, arr, groups_indices)
    arg = arg_top_max(means, num_outliers)
    result = [
        (groups_indices[idx][0], groups_indices[idx][-1])
        for idx, _ in enumerate(means)
        if idx in arg
    ]

    return result


def scores_to_intervals(scores, high_percentile=90, num_selected=3):
    [low, high] = list(indices_for_percentile(scores, low=0, high=high_percentile))
    arr_selected = np.nonzero(scores >= high)[0]
    groups_indices = consecutive_indices(arr_selected)
    intervals_all = get_groups_extremities_all(
        scores, groups_indices, func=np.mean, num_outliers=num_selected,
    )
    result = [(a, b) for a, b in intervals_all if b > a]

    return result


# Chunkers
def simple_chunker(it, chk_size: int = DFLT_CHK_SIZE):
    return fixed_step_chunker(it=it, chk_size=chk_size, chk_step=chk_size)


DFLT_CHUNKER = simple_chunker
# Featurizers
DFLT_FEATURIZER = lambda chk: np.abs(np.fft.rfft(chk))


def simple_featurizer(chks):
    fvs = np.array(list(map(DFLT_FEATURIZER, chks)))
    return fvs


DFLT_PIPELINE = LineParametrized(simple_chunker, simple_featurizer)

featurizer = DFLT_FEATURIZER
chunker = simple_chunker
WaveForm = Iterable[int]


def clean_dict(kwargs):
    result = {k: v for k, v in kwargs.items() if v != ''}
    return result


def store_to_key_fvs(wf_store, chunker=DFLT_CHUNKER, featurizer=DFLT_FEATURIZER):
    wf_items = wf_store.items()
    key_chk_tuples = fanout_and_flatten(wf_items, chunker, 1)
    featurizer_iter = partial(apply_func_to_index, apply_to_idx=1, func=featurizer)
    yield from map(featurizer_iter, key_chk_tuples)


def key_fvs_to_tag_fvs(key_fvs, annots_df):
    func = partial(key_to_tag_from_annots, annots_df=annots_df)
    tagger = partial(apply_func_to_index, apply_to_idx=0, func=func)

    yield from map(tagger, key_fvs)


def key_to_tag_from_annots(key, annots_df):
    tag = annots_df['tag'][annots_df['key'] == key].values[0]
    return tag


def mk_Xy(tag_fv_iterator):
    y, X = zip(*list(tag_fv_iterator))
    return np.array(X), y


DFLT_CHAIN = (({'type': 'pca', 'args': {'n_components': 5}},),)


def preprocess(X_train, n_components=5):

    from shaded.chained_spectral_projector import learn_chain_proj_matrix

    chain = DFLT_CHAIN
    chain['args']['n_components'] = n_components
    proj_matrix = learn_chain_proj_matrix(X_train, chain=chain)
    X_train_proj = np.dot(X_train, proj_matrix)

    return proj_matrix, X_train_proj


def simple_model(tagged_data):
    sound, tag = tagged_data
    if not isinstance(sound, str):
        sound = sound.getvalue()

    arr = sf.read(BytesIO(sound), dtype='int16')[0]

    wfs = np.array(arr)
    st.write(f'wfs = {wfs[:200]}')
    chks = list(chunker(wfs, chk_size=DFLT_CHK_SIZE))
    fvs = np.array(list(map(featurizer, chks)))
    model = Stroll(n_centroids=50)
    model.fit(X=fvs)
    scores = model.score_samples(X=fvs)
    return scores


def assert_dims(wfs):
    if wfs.ndim >= 2:
        wfs = wfs[:, 0]
    return wfs


def tagged_sound_to_array(train_audio: WaveForm, tag: str):
    sound, tag = train_audio, tag
    if not isinstance(sound, bytes):
        sound = sound.getvalue()

    arr = sf.read(BytesIO(sound), dtype='int16')[0]
    return arr, tag


def tagged_sounds_to_single_array(train_audio: List[WaveForm], tag: str):
    sounds, tag = train_audio, tag
    result = []
    for sound in sounds:
        # if not isinstance(sound, bytes):
        sound = sound.getvalue()
        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        result.append(arr)
    # print(np.hstack(result))
    return np.hstack(result).reshape(-1, 1), tag



import streamlit as st
from meshed import DAG
from front.types import FrontElementName
from front.elements import FrontComponentBase, TextInputBase
from typing import List, Any

WaveForm = List[int]


class SimpleList(FrontComponentBase):
    def __init__(
        self,
        arr: WaveForm,
        name: FrontElementName = None,
        use_container_width: bool = False,
    ):
        super().__init__(arr=arr, name=name)
        self.use_container_width = use_container_width

    def render(self):
        with st.expander(self.name, True):
            arr: WaveForm = self.arr
            st.write(arr.len)


class SimpleText(TextInputBase):
    def __init__(
        self,
        obj: str,
        name: FrontElementName = None,
        use_container_width: bool = False,
        input_key: str = None,
        init_value=None,
    ):
        super().__init__(obj=obj, name=name, input_key=input_key)
        self.use_container_width = use_container_width

    def render(self):
        # with st.expander(self.name, True): No nesting of expanders
        input_key: str = self.input_key
        st.write(f'length of text entered= {len(input_key)}')


class SimpleDataframe(FrontComponentBase):
    def __init__(
        self,
        obj: str,
        name: FrontElementName = None,
        use_container_width: bool = False,
        input_key: str = None,
        init_value=None,
    ):
        super().__init__(obj=obj, name=name, input_key=input_key)
        self.use_container_width = use_container_width

    def render(self):
        # with st.expander(self.name, True): No nesting of expanders
        input_key: str = self.input_key
        st.write(f'length of text entered= {len(input_key)}')



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from typing import Callable, Iterable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from i2 import Pipe, Sig
from front.crude import Crudifier, prepare_for_crude_dispatch
from lined import LineParametrized
import numpy as np
from meshed import DAG
from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    PipelineMaker,
)
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
from functools import partial
from front.crude import crudify_based_on_names, prepare_for_crude_dispatch
from collections import defaultdict
import streamlit as st

import soundfile as sf
from io import BytesIO
from plunk.sb.front_demo.user_story1.components.components import (
    AudioArrayDisplay,
    GraphOutput,
    ArrayPlotter,
    ArrayWithIntervalsPlotter,
)
from plunk.sb.front_demo.user_story1.utils.tools import (
    DFLT_FEATURIZER,
    DFLT_CHK_SIZE,
    chunker,
)
from typing import List


def simple_chunker(wfs, chk_size: int = DFLT_CHK_SIZE):
    return list(chunker(wfs, chk_size=chk_size))


def simple_featurizer(chks):
    fvs = np.array(list(map(DFLT_FEATURIZER, chks)))
    return fvs


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_store=None,
    learned_models=None,
    models_scores=None,
):
    from plunk.sb.front_demo.user_story1.utils.funcs import (
        upload_sound,
        # display_tag_sound,
        mk_step,
        mk_pipeline,
        learn_outlier_model,
        apply_fitted_model,
        # exec_pipeline,
        visualize_pipeline,
        visualize_scores,
    )

    if not b.mall():
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'chunker'  # TODO make this dynamic

    steps_store = steps_store or steps
    data_store = data_store or data
    pipelines_store = pipelines_store or pipelines

    general_crudifier = partial(
        crudify_based_on_names,
        param_to_mall_map={
            'step_factory': 'step_factories',
            'tagged_data': 'sound_output',
            'preprocess_pipeline': 'pipelines',
            'tagged_data': 'sound_output',
            'preprocess_pipeline': 'pipelines',
            'fitted_model': 'learned_models',
            'pipeline': 'pipelines_store',
            'scores': 'models_scores',
        },
        output_store={
            mk_step: 'steps_store',
            mk_pipeline: 'pipelines_store',
            learn_outlier_model: 'learned_models',
            apply_fitted_model: 'models_scores',
            upload_sound: 'sound_output',
        },
        crudifier=partial(prepare_for_crude_dispatch, mall=mall),
    )

    [
        upload_sound,
        # display_tag_sound,
        mk_step,
        mk_pipeline,
        learn_outlier_model,
        apply_fitted_model,
        # exec_pipeline,
        visualize_pipeline,
        visualize_scores,
    ] = map(
        general_crudifier,
        [
            upload_sound,
            # display_tag_sound,
            mk_step,
            mk_pipeline,
            learn_outlier_model,
            apply_fitted_model,
            # exec_pipeline,
            visualize_pipeline,
            visualize_scores,
        ],
    )

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v == step][0]

    def get_selected_step_factory_sig():
        selected_step_factory = mall['step_factories'].get(
            b.selected_step_factory.get()
        )
        if selected_step_factory:
            return Sig(selected_step_factory)

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'accept_multiple_files': True,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav files loaded successfully.',
                    },
                },
            },
            'display_tag_sound': {
                'execution': {
                    # 'inputs': {
                    #     'result': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['sound_output'],
                    #     },
                    # },
                    'output': {ELEMENT_KEY: AudioArrayDisplay,},
                },
            },
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {'func_sig': get_selected_step_factory_sig},
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': list(mall[steps].values()),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'exec_pipeline': {
                NAME_KEY: 'Pipeline Executor',
                'execution': {
                    'inputs': {
                        'pipeline': {'value': b.selected_pipeline,},
                        # 'tagged_data': {
                        #     ELEMENT_KEY: SelectBox,
                        #     'options': mall['sound_output'],
                        # },
                    }
                },
            },
            'visualize_pipeline': {
                NAME_KEY: 'Pipeline Visualization',
                'execution': {
                    'inputs': {'pipeline': {'value': b.selected_pipeline,},},
                    'output': {
                        ELEMENT_KEY: GraphOutput,
                        NAME_KEY: 'Flow',
                        'use_container_width': True,
                    },
                },
            },
            'visualize_scores': {
                NAME_KEY: 'Scores Visualization',
                'execution': {
                    # 'inputs': {
                    #     'scores': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['models_scores'],
                    #     },
                    # },
                    'output': {ELEMENT_KEY: ArrayWithIntervalsPlotter,},
                },
            },
            'simple_model': {
                NAME_KEY: 'Learn model',
                'execution': {
                    # 'inputs': {
                    #     'tagged_data': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['sound_output'],
                    #     },
                    #     'preprocess_pipeline': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['pipelines'],
                    #     },
                    # },
                    'output': {ELEMENT_KEY: ArrayPlotter,},
                },
            },
            'apply_fitted_model': {
                NAME_KEY: 'Apply model',
                'execution': {
                    # 'inputs': {
                    #     'tagged_data': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['sound_output'],
                    #     },
                    #     'preprocess_pipeline': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['pipelines'],
                    #     },
                    #     'learned_model': {
                    #         ELEMENT_KEY: SelectBox,
                    #         'options': mall['learned_models'],
                    #     },
                    # },
                    'output': {ELEMENT_KEY: ArrayPlotter,},
                },
            },
        },
    }

    funcs = [
        upload_sound,
        # display_tag_sound,
        mk_step,
        mk_pipeline,
        learn_outlier_model,
        apply_fitted_model,
        # exec_pipeline,
        visualize_pipeline,
        visualize_scores,
    ]
    app = mk_app(funcs, config=config)

    return app


if __name__ == '__main__':

    mall = dict(
        # Factory Input Stores
        sound_output=dict(),
        step_factories=dict(),
        # Output Store
        data=dict(),
        steps=dict(),
        steps_store=dict(),
        pipelines_store=dict(),
        pipelines=dict(),
        exec_outputs=dict(),
        learned_models=dict(),
        models_scores=dict(),
    )

    # crudifier = partial(prepare_for_crude_dispatch, mall=mall)

    step_factories = dict(
        # ML
        chunker=FuncFactory(simple_chunker),
        featurizer=FuncFactory(simple_featurizer),
    )

    mall['step_factories'] = step_factories

    app = mk_pipeline_maker_app_with_mall(
        mall, step_factories='step_factories', steps='steps', pipelines='pipelines'
    )

    app()



"""
An app to take care of the initial 'sourcing' part of the data prep of audio ML
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from typing import Callable, Iterable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from i2 import Pipe, Sig
from front.crude import Crudifier
from slang.chunkers import fixed_step_chunker
import streamlit as st
from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    KwargsInput,
    PipelineMaker,
)
from slang import fixed_step_chunker, mk_chk_fft

from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep2 import (
    # DFLT_WF_PATH,
    # DFLT_ANNOT_PATH,
    data_from_wav_folder,
    data_from_csv,
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)


def chunker(it, chk_size):
    return fixed_step_chunker(it=it, chk_size=chk_size, chk_step=chk_size)


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
):
    # TODO: Like to not have this binder logic involving streamlit state here! Contain it!
    if not b.mall():
        # TODO: Maybe it's here that we need to use know.malls.mk_mall?
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'data_loader'

    crudifier = partial(Crudifier, mall=mall)

    steps_store = steps_store or steps
    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    def mk_step(step_factory: Callable, kwargs: dict):
        return partial(step_factory, **kwargs)

        # def mk_step(func): crudified to be sourced from funcfactory store
        mk_app(choice)

    #
    @crudifier(
        # TODO: Want to be able to do this and this only to have the effect
        # param_to_mall_map=dict(steps=steps),
        output_store=pipelines_store
    )
    def mk_pipeline(steps: Iterable[Callable]):
        print(f'{steps}')
        return Pipe(*steps)

    @crudifier(
        # TODO: Does this work if pipelines_store is a mapping instead of a string?
        param_to_mall_map=dict(pipeline=pipelines_store),
        output_store='exec_outputs',
    )
    def exec_pipeline(pipeline: Callable, kwargs):
        print(type(pipeline), Sig(pipeline))
        return pipeline(**kwargs)

    # NOTE: Just ideas. call_func not used:
    # @crudifier(
    #     param_to_mall_map=dict(pipeline='pipelines'),
    #     # output_store='exec_outputs'
    # )
    def call_func(func: Callable, *args, **kwargs):
        return func(*args, **kwargs)

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v == step][0]

    def get_selected_pipeline_sig():
        if not b.selected_pipeline():
            return Sig()
        return Sig(mall[pipelines][b.selected_pipeline()])

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall[step_factories],
                            'value': b.selected_step_factory,
                        },
                        'kwargs': {
                            ELEMENT_KEY: KwargsInput,
                            'func_sig': Sig(
                                mall[step_factories][b.selected_step_factory()]
                            ),
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': list(mall[steps].values()),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'exec_pipeline': {
                NAME_KEY: 'Pipeline Executor',
                'execution': {
                    'inputs': {
                        'pipeline': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall[pipelines],
                            'value': b.selected_pipeline,
                        },
                        'kwargs': {
                            ELEMENT_KEY: KwargsInput,
                            'func_sig': get_selected_pipeline_sig(),
                        },
                    }
                },
            },
        },
    }

    funcs = [mk_step, mk_pipeline, exec_pipeline]
    app = mk_app(funcs, config=config)

    return app


if __name__ == '__main__':
    # TODO: Try with know.malls.mk_mall:
    mall = dict(
        step_factories=dict(
            # Source Readers
            data_loader=FuncFactory(data_from_wav_folder),
            # data_loader=FuncFactory(foo),
            # Chunkers
            chunker=FuncFactory(chunker),
        ),
        steps=dict(),
        pipelines=dict(),
        exec_outputs=dict(),
    )

    app = mk_pipeline_maker_app_with_mall(
        mall, step_factories='step_factories', steps='steps', pipelines='pipelines'
    )

    app()



import os
import streamlit.components.v1 as components


def mk_element_factory(element_name):
    parent_dir = os.path.dirname(os.path.abspath(__file__))
    build_dir = os.path.join(parent_dir, element_name)
    return components.declare_component(element_name, path=build_dir)



import os
import time
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from io import BytesIO
from pathlib import Path
from typing import Iterable
import matplotlib.pyplot as plt
import soundfile as sf
import streamlit as st
from dol import Files
from dol.appendable import appendable
from front import APP_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY, RENDERING_KEY
from front.crude import Crudifier, prepare_for_crude_dispatch
from front.elements import DEFAULT_INPUT_KEY, OutputBase
from meshed import DAG, code_to_dag

from streamlitfront import binder as b
from streamlitfront import mk_app
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
    SelectBox,
    TextInput,
)
from streamlitfront.examples.util import Graph

# ============ BACKEND ============
WaveForm = Iterable[int]

# rootdir = os.path.join(Path('~').expanduser(), '.front', 'edge_impluse_like')
# os.makedirs(rootdir, exist_ok=True)

# def tagged_timestamped_kv(tag_and_val_item, default_tag='untagged'):
#     utc_seconds_timestamp = int(time.time())
#     tag, val = tag_and_val_item
#     key = f"{utc_seconds_timestamp}-{tag}"

#     return key, val

# tagged_wf_store = appendable(Files, item2kv=tagged_timestamped_kv)
if not b.mall():
    b.mall = dict(tagged_wf=dict(), dummy_store=dict(),)
mall = b.mall()
crudifier = partial(Crudifier, mall=mall)


def auto_namer(*, arguments):
    return '_'.join(map(str, arguments.values()))


@crudifier(output_store='tagged_wf', auto_namer=auto_namer)
def tag_wf(wf: WaveForm, tag: str):
    return (wf, tag)


@crudifier(
    param_to_mall_map=dict(x='tagged_wf'),
    output_store='dummy_store',
    auto_namer=auto_namer,
)
def get_tagged_wf(x):
    return x


# ============ END BACKEND ============


# ============ FRONTEND ============
def timestamped_kv(value):
    utc_seconds_timestamp = str(int(time.time() * 1000))
    return utc_seconds_timestamp, value


AppendableFiles = appendable(Files, item2kv=timestamped_kv, return_keys=True)


rootdir = os.path.join(Path('~').expanduser(), '.front', 'edge_impulse_like')
Path(rootdir).mkdir(parents=True, exist_ok=True)


@dataclass
class AudioPersister(AudioRecorder):
    save_dir: str = None

    def __post_init__(self):
        super().__post_init__()
        self.store = AppendableFiles(self.save_dir)
        # self.store = Files(self.save_dir) if self.save_dir else None

    def render(self):
        audio_data = super().render()
        if audio_data and self.store is not None:
            print('YEAH')
            key = self.store.append(audio_data)
            return os.path.join(self.save_dir, key)
            # self.store.append(audio_data)
        return audio_data


class TaggedAudioPlayer(OutputBase):
    def render(self):
        sound, tag = self.output
        if not isinstance(sound, str):
            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        tab1, tab2 = st.tabs(['Audio Player', 'Waveform'])
        with tab1:
            st.audio(sound)
        with tab2:
            fig, ax = plt.subplots(figsize=(15, 5))
            ax.plot(arr, label=f'Tag={tag}')
            ax.legend()
            st.pyplot(fig)
            # st.write(arr[:10])


get_data_description = '''
Feed the system with wave forms from wav files or directly \
from your microphone, and tag them.
'''

config_ = {
    APP_KEY: {'title': 'Edge-Impulse-like App'},
    RENDERING_KEY: {
        'tag_wf': {
            NAME_KEY: 'Get Data',
            'description': {'content': get_data_description},
            'execution': {
                'inputs': {
                    'wf': {
                        ELEMENT_KEY: MultiSourceInput,
                        NAME_KEY: 'Wave Form',
                        'From a file': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'display_label': False,
                        },
                        'From the microphone': {
                            ELEMENT_KEY: AudioPersister,
                            'save_dir': rootdir,
                        },
                    }
                }
            },
        },
        'get_tagged_wf': {
            NAME_KEY: 'Data Explorer',
            'description': {
                'content': '''Explore the existing tagged wave forms and play them.'''
            },
            'execution': {
                'inputs': {
                    'x': {ELEMENT_KEY: SelectBox, 'options': mall['tagged_wf'],}
                },
                'output': {ELEMENT_KEY: TaggedAudioPlayer,},
                'auto_submit': True,
            },
        },
        DAG: {'graph': {ELEMENT_KEY: Graph, NAME_KEY: 'Flow',}},
        Callable: {
            'execution': {
                'inputs': {
                    'save_name': {ELEMENT_KEY: TextInput, NAME_KEY: 'Save output as',}
                }
            }
        },
    },
}
# ============ END FRONTEND ============

if __name__ == '__main__':
    app = mk_app([tag_wf, get_tagged_wf], config=config_)
    app()



import streamlit as st
from io import StringIO

# def func():
#    st.write()

uploaded_file = st.file_uploader('Choose a file')
if uploaded_file is not None:
    # To read file as bytes:
    # bytes_data = uploaded_file.getvalue()
    # st.write(bytes_data)

    # To convert to a string based IO:
    # stringio = StringIO(uploaded_file.getvalue().decode("utf-8"))
    # st.write(stringio)

    # To read file as string:
    # string_data = stringio.read()
    # st.write(string_data)

    # Can be used wherever a "file-like" object is accepted:
    # dataframe = pd.read_csv(uploaded_file)
    # st.write(dataframe)
    st.write(vars(uploaded_file))



from distutils import errors
from distutils.log import error
import streamlit as st
import pandas as pd
import numpy as np
import altair as alt
from itertools import cycle

from st_aggrid import GridOptionsBuilder, AgGrid, GridUpdateMode, DataReturnMode, JsCode

np.random.seed(42)


@st.cache(allow_output_mutation=True)
def fetch_data(samples):
    deltas = cycle(
        [
            pd.Timedelta(weeks=-2),
            pd.Timedelta(days=-1),
            pd.Timedelta(hours=-1),
            pd.Timedelta(0),
            pd.Timedelta(minutes=5),
            pd.Timedelta(seconds=10),
            pd.Timedelta(microseconds=50),
            pd.Timedelta(microseconds=10),
        ]
    )
    dummy_data = {
        'date_time_naive': pd.date_range('2021-01-01', periods=samples),
        'apple': np.random.randint(0, 100, samples) / 3.0,
        'banana': np.random.randint(0, 100, samples) / 5.0,
        'chocolate': np.random.randint(0, 100, samples),
        'group': np.random.choice(['A', 'B'], size=samples),
        'date_only': pd.date_range('2020-01-01', periods=samples).date,
        'timedelta': [next(deltas) for i in range(samples)],
        'date_tz_aware': pd.date_range(
            '2022-01-01', periods=samples, tz='Asia/Katmandu'
        ),
    }
    return pd.DataFrame(dummy_data)


# Example controlers
st.sidebar.subheader('St-AgGrid example options')

sample_size = st.sidebar.number_input('rows', min_value=10, value=30)
grid_height = st.sidebar.number_input(
    'Grid height', min_value=200, max_value=800, value=300
)

return_mode = st.sidebar.selectbox(
    'Return Mode', list(DataReturnMode.__members__), index=1
)
return_mode_value = DataReturnMode.__members__[return_mode]

update_mode = st.sidebar.selectbox(
    'Update Mode',
    list(GridUpdateMode.__members__),
    index=len(GridUpdateMode.__members__) - 1,
)
update_mode_value = GridUpdateMode.__members__[update_mode]

# enterprise modules
enable_enterprise_modules = st.sidebar.checkbox('Enable Enterprise Modules')
if enable_enterprise_modules:
    enable_sidebar = st.sidebar.checkbox('Enable grid sidebar', value=False)
else:
    enable_sidebar = False

# features
fit_columns_on_grid_load = st.sidebar.checkbox('Fit Grid Columns on Load')

enable_selection = st.sidebar.checkbox('Enable row selection', value=True)
if enable_selection:
    st.sidebar.subheader('Selection options')
    selection_mode = st.sidebar.radio('Selection Mode', ['single', 'multiple'], index=1)

    use_checkbox = st.sidebar.checkbox('Use check box for selection', value=True)
    if use_checkbox:
        groupSelectsChildren = st.sidebar.checkbox(
            'Group checkbox select children', value=True
        )
        groupSelectsFiltered = st.sidebar.checkbox(
            'Group checkbox includes filtered', value=True
        )

    if (selection_mode == 'multiple') & (not use_checkbox):
        rowMultiSelectWithClick = st.sidebar.checkbox(
            'Multiselect with click (instead of holding CTRL)', value=False
        )
        if not rowMultiSelectWithClick:
            suppressRowDeselection = st.sidebar.checkbox(
                'Suppress deselection (while holding CTRL)', value=False
            )
        else:
            suppressRowDeselection = False
    st.sidebar.text('___')

enable_pagination = st.sidebar.checkbox('Enable pagination', value=False)
if enable_pagination:
    st.sidebar.subheader('Pagination options')
    paginationAutoSize = st.sidebar.checkbox('Auto pagination size', value=True)
    if not paginationAutoSize:
        paginationPageSize = st.sidebar.number_input(
            'Page size', value=5, min_value=0, max_value=sample_size
        )
    st.sidebar.text('___')

df = fetch_data(sample_size)

# Infer basic colDefs from dataframe types
gb = GridOptionsBuilder.from_dataframe(df)

# customize gridOptions
gb.configure_default_column(
    groupable=True, value=True, enableRowGroup=True, aggFunc='sum', editable=True
)
gb.configure_column(
    'date_only',
    type=['dateColumnFilter', 'customDateTimeFormat'],
    custom_format_string='yyyy-MM-dd',
    pivot=True,
)
gb.configure_column(
    'date_tz_aware',
    type=['dateColumnFilter', 'customDateTimeFormat'],
    custom_format_string='yyyy-MM-dd HH:mm zzz',
    pivot=True,
)

gb.configure_column(
    'apple',
    type=['numericColumn', 'numberColumnFilter', 'customNumericFormat'],
    precision=2,
    aggFunc='sum',
)
gb.configure_column(
    'banana',
    type=['numericColumn', 'numberColumnFilter', 'customNumericFormat'],
    precision=1,
    aggFunc='avg',
)
gb.configure_column(
    'chocolate',
    type=['numericColumn', 'numberColumnFilter', 'customCurrencyFormat'],
    custom_currency_symbol='R$',
    aggFunc='max',
)

# configures last row to use custom styles based on cell's value, injecting JsCode on components front end
cellsytle_jscode = JsCode(
    '''
function(params) {
    if (params.value == 'A') {
        return {
            'color': 'white',
            'backgroundColor': 'darkred'
        }
    } else {
        return {
            'color': 'black',
            'backgroundColor': 'white'
        }
    }
};
'''
)
gb.configure_column('group', cellStyle=cellsytle_jscode)

if enable_sidebar:
    gb.configure_side_bar()

if enable_selection:
    gb.configure_selection(selection_mode)
    if use_checkbox:
        gb.configure_selection(
            selection_mode,
            use_checkbox=True,
            groupSelectsChildren=groupSelectsChildren,
            groupSelectsFiltered=groupSelectsFiltered,
        )
    if (selection_mode == 'multiple') & (not use_checkbox):
        gb.configure_selection(
            selection_mode,
            use_checkbox=False,
            rowMultiSelectWithClick=rowMultiSelectWithClick,
            suppressRowDeselection=suppressRowDeselection,
        )

if enable_pagination:
    if paginationAutoSize:
        gb.configure_pagination(paginationAutoPageSize=True)
    else:
        gb.configure_pagination(
            paginationAutoPageSize=False, paginationPageSize=paginationPageSize
        )

gb.configure_grid_options(domLayout='normal')
gridOptions = gb.build()

# Display the grid
st.header('Streamlit Ag-Grid')
st.markdown(
    '''
    AgGrid can handle many types of columns and will try to render the most human readable way.  
    On editions, grid will fallback to string representation of data, DateTime and TimeDeltas are converted to ISO format.
    Custom display formating may be applied to numeric fields, but returned data will still be numeric.
'''
)

grid_response = AgGrid(
    df,
    gridOptions=gridOptions,
    height=grid_height,
    width='100%',
    data_return_mode=return_mode_value,
    update_mode=update_mode_value,
    fit_columns_on_grid_load=fit_columns_on_grid_load,
    allow_unsafe_jscode=True,  # Set it to True to allow jsfunction to be injected
    enable_enterprise_modules=enable_enterprise_modules,
)

df = grid_response['data']
selected = grid_response['selected_rows']
selected_df = pd.DataFrame(selected).apply(pd.to_numeric, errors='coerce')


with st.spinner('Displaying results...'):
    # displays the chart
    chart_data = df.loc[:, ['apple', 'banana', 'chocolate']].assign(source='total')

    if not selected_df.empty:
        selected_data = selected_df.loc[:, ['apple', 'banana', 'chocolate']].assign(
            source='selection'
        )
        chart_data = pd.concat([chart_data, selected_data])

    chart_data = pd.melt(
        chart_data, id_vars=['source'], var_name='item', value_name='quantity'
    )
    # st.dataframe(chart_data)
    chart = (
        alt.Chart(data=chart_data)
        .mark_bar()
        .encode(
            x=alt.X('item:O'),
            y=alt.Y('sum(quantity):Q', stack=False),
            color=alt.Color('source:N', scale=alt.Scale(domain=['total', 'selection'])),
        )
    )

    st.header('Component Outputs - Example chart')
    st.markdown(
        '''
    This chart is built with data returned from the grid. rows that are selected are also identified.
    Experiment selecting rows, group and filtering and check how the chart updates to match.
    '''
    )

    st.altair_chart(chart, use_container_width=True)

    st.subheader('Returned grid data:')
    # returning as HTML table bc streamlit has issues when rendering dataframes with timedeltas:
    # https://github.com/streamlit/streamlit/issues/3781
    st.markdown(grid_response['data'].to_html(), unsafe_allow_html=True)

    st.subheader('grid selection:')
    st.write(grid_response['selected_rows'])

    st.header('Generated gridOptions')
    st.markdown(
        '''
        All grid configuration is done thorugh a dictionary passed as ```gridOptions``` parameter to AgGrid call.
        You can build it yourself, or use ```gridOptionBuilder``` helper class.  
        Ag-Grid documentation can be read [here](https://www.ag-grid.com/documentation)
    '''
    )
    st.write(gridOptions)



import time
import pandas as pd
import plotly.express as px
import streamlit as st

st.set_page_config(
    page_title='Stream a dataframe', layout='wide',
)

# read csv from local folder
dataset_path = 'data/accel_data.csv'


@st.experimental_memo
def get_data(filepath) -> pd.DataFrame:
    return pd.read_csv(filepath, sep=';')


df = get_data(dataset_path)
min_data = df['ACC_Y'].min()
max_data = df['ACC_Y'].max()

# dashboard title
st.title('Stream a dataframe')


# creating a single-element container
placeholder = st.empty()

# simulate live data
for seconds in range(200):

    vals = df['ACC_Y'].iloc[seconds : seconds + 30].values
    d = {'ACC_Y': vals}
    with placeholder.container():
        fig = px.line(
            data_frame=pd.DataFrame(d),
            x=range(30),
            y='ACC_Y',
            range_y=[min_data, max_data],
        )
        st.write(fig)

        time.sleep(0.1)



from distutils.command.config import config
import streamlit as st
from streamlitfront.base import mk_app
from front.elements import InputComponentFlag, InputBase
from streamlitfront.elements.elements import implement_component_with_init_value


FileInput = implement_component_with_init_value(InputBase, st.file_uploader)


def foo(a: int = 1, b: int = 2, c=3):
    """This is foo. It computes something"""
    return (a * b) + c


def bar(x, greeting='hello'):
    """bar greets its input"""
    return f'{greeting} {x}'


def confuser(a: int, x: float = 3.14):
    return (a ** 2) * x


def proportion(x: int = 100, p: float = 0.5):
    return x * p


app = mk_app(
    [foo, bar, confuser, proportion],
    config={
        'app': {'title': 'My app'},
        # 'obj': {
        #     'bindings': {
        #         'Foo.a': 'Proportion.x'
        #     }
        # }
        'rendering': {
            'Proportion': {
                'inputs': {'p': {'component': InputComponentFlag.FLOAT_SLIDER,}},
            }
        },
    },
)
app()



"""
Investigate a bug found by Christian
kwargs are heavily modified by a dag
"""

from meshed import DAG, FuncNode
from typing import Union


def maybe_first(items):
    return next(iter(items), None)


def name_of_var_kw_argument(sig):
    var_kw_list = [param.name for param in sig.params if param.kind == VK]
    result = maybe_first(var_kw_list)
    return result


def map_action_on_cond(kvs, cond, expand):
    for k, v in kvs:
        if cond(
            k
        ):  # make a conditional on (k,v), use type KV, Iterable[KV], expand:KV -> Iterable[KV]
            yield from expand(v)  # expand should result in (k,v)
        else:
            yield k, v


def flatten_if_var_kw(kvs, var_kw_name):
    cond = lambda k: k == var_kw_name
    expand = lambda k: k.items()

    return map_action_on_cond(kvs, cond, expand)


def kwargs_from_args_and_kwargs(
    self,
    args,
    kwargs,
    *,
    apply_defaults=False,
    allow_partial=False,
    allow_excess=False,
    ignore_kind=False,
    debug=False,
):
    """Extracts a dict of input argument values for target signature, from args
    and kwargs.

    When you need to manage how the arguments of a function are specified,
    you need to take care of
    multiple cases depending on whether they were specified as positional arguments
    (`args`) or keyword arguments (`kwargs`).

    The `kwargs_from_args_and_kwargs` (and it's sorta-inverse inverse,
    `args_and_kwargs_from_kwargs`)
    are there to help you manage this.

    If you could rely on the the fact that only `kwargs` were given it would
    reduce the complexity of your code.
    This is why we have the `all_pk_signature` function in `signatures.py`.

    We also need to have a means to make a `kwargs` only from the actual `(*args,
    **kwargs)` used at runtime.
    We have `Signature.bind` (and `bind_partial`) for that.

    But these methods will fail if there is extra stuff in the `kwargs`.
    Yet sometimes we'd like to have a `dict` that services several functions that
    will extract their needs from it.

    That's where  `Sig.extract_kwargs(*args, **kwargs)` is needed.
    :param args: The args the function will be called with.
    :param kwargs: The kwargs the function will be called with.
    :param apply_defaults: (bool) Whether to apply signature defaults to the
    non-specified argument names
    :param allow_partial: (bool) True iff you want to allow partial signature
    fulfillment.
    :param allow_excess: (bool) Set to True iff you want to allow extra kwargs
    items to be ignored.
    :param ignore_kind: (bool) Set to True iff you want to ignore the position and
    keyword only kinds,
        in order to be able to accept args and kwargs in such a way that there can
        be cross-over
        (args that are supposed to be keyword only, and kwargs that are supposed
        to be positional only)
    :return: An {argname: argval, ...} dict

    See also the sorta-inverse of this function: args_and_kwargs_from_kwargs

    >>> def foo(w, /, x: float, y="YY", *, z: str = "ZZ"):
    ...     ...
    >>> sig = Sig(foo)
    >>> assert (
    ...     sig.kwargs_from_args_and_kwargs((11, 22, "you"), dict(z="zoo"))
    ...     == sig.kwargs_from_args_and_kwargs((11, 22), dict(y="you", z="zoo"))
    ...     == {"w": 11, "x": 22, "y": "you", "z": "zoo"}
    ... )

    By default, `apply_defaults=False`, which will lead to only get those
    arguments you input.

    >>> sig.kwargs_from_args_and_kwargs(args=(11,), kwargs={"x": 22})
    {'w': 11, 'x': 22}

    But if you specify `apply_defaults=True` non-specified non-require arguments
    will be returned with their defaults:

    >>> sig.kwargs_from_args_and_kwargs(
    ...     args=(11,), kwargs={"x": 22}, apply_defaults=True
    ... )
    {'w': 11, 'x': 22, 'y': 'YY', 'z': 'ZZ'}

    By default, `ignore_excess=False`, so specifying kwargs that are not in the
    signature will lead to an exception.

    >>> sig.kwargs_from_args_and_kwargs(
    ...     args=(11,), kwargs={"x": 22, "not_in_sig": -1}
    ... )
    Traceback (most recent call last):
        ...
    TypeError: Got unexpected keyword arguments: not_in_sig

    Specifying `allow_excess=True` will ignore such excess fields of kwargs.
    This is useful when you want to source several functions from a same dict.

    >>> sig.kwargs_from_args_and_kwargs(
    ...     args=(11,), kwargs={"x": 22, "not_in_sig": -1}, allow_excess=True
    ... )
    {'w': 11, 'x': 22}

    On the other side of `ignore_excess` you have `allow_partial` that will allow
    you, if
    set to `True`, to underspecify the params of a function (in view of being
    completed later).

    >>> sig.kwargs_from_args_and_kwargs(args=(), kwargs={"x": 22})
    Traceback (most recent call last):
      ...
    TypeError: missing a required argument: 'w'

    But if you specify `allow_partial=True`...

    >>> sig.kwargs_from_args_and_kwargs(
    ...     args=(), kwargs={"x": 22}, allow_partial=True
    ... )
    {'x': 22}

    That's a lot of control (eight combinations total), but not everything is
    controllable here:
    Position only and keyword only kinds need to be respected:

    >>> sig.kwargs_from_args_and_kwargs(args=(1, 2, 3, 4), kwargs={})
    Traceback (most recent call last):
      ...
    TypeError: too many positional arguments
    >>> sig.kwargs_from_args_and_kwargs(args=(), kwargs=dict(w=1, x=2, y=3, z=4))
    Traceback (most recent call last):
      ...
    TypeError: 'w' parameter is positional only, but was passed as a keyword

    But if you want to ignore the kind of parameter, just say so:

    >>> sig.kwargs_from_args_and_kwargs(
    ...     args=(1, 2, 3, 4), kwargs={}, ignore_kind=True
    ... )
    {'w': 1, 'x': 2, 'y': 3, 'z': 4}
    >>> sig.kwargs_from_args_and_kwargs(
    ...     args=(), kwargs=dict(w=1, x=2, y=3, z=4), ignore_kind=True
    ... )
    {'w': 1, 'x': 2, 'y': 3, 'z': 4}
    """
    no_var_kw = not self.has_var_keyword
    # no_var_kw = True

    if ignore_kind:
        sig = self.normalize_kind(
            # except_kinds=frozenset()
        )
    else:
        sig = self

    # no_var_kw = not sig.has_var_keyword TODOD check this
    if no_var_kw:  # has no var keyword kinds
        sig_relevant_kwargs = {
            name: kwargs[name] for name in sig if name in kwargs
        }  # take only what you need
    else:
        sig_relevant_kwargs = kwargs  # take all the kwargs
    binder = sig.bind_partial if allow_partial else sig.bind
    if not self.has_var_positional and allow_excess:
        max_allowed_num_of_posisional_args = sum(k <= PK for k in self.kinds.values())
        args = args[:max_allowed_num_of_posisional_args]

    b = binder(*args, **sig_relevant_kwargs)
    if apply_defaults:
        b.apply_defaults()

    if no_var_kw and not allow_excess:  # don't ignore excess kwargs
        excess = kwargs.keys() - b.arguments
        if excess:
            excess_str = ', '.join(excess)
            raise TypeError(f'Got unexpected keyword arguments: {excess_str}')

    if debug:
        var_kw_name = name_of_var_kw_argument(self)

        kvs = b.arguments.items()

        flattened_kvs = flatten_if_var_kw(kvs, var_kw_name)
        result = dict(flattened_kvs)

    else:
        result = dict(b.arguments)
    return result


def args_and_kwargs_from_kwargs(
    self,
    kwargs,
    apply_defaults=False,
    allow_partial=False,
    allow_excess=False,
    ignore_kind=False,
    args_limit: Union[int, None] = 0,
):
    """Extract args and kwargs such that func(*args, **kwargs) can be called,
    where func has instance's signature.

    :param kwargs: The {argname: argval,...} dict to process
    :param args_limit: How "far" in the params should args (positional arguments)
        be searched for.
        - args_limit==0: Take the minimum number possible of args (positional
            arguments). Only those that are position only or before a var-positional.
        - args_limit is None: Take the maximum number of args (positional arguments).
            The only kwargs (keyword arguments) you should have are keyword-only
            and var-keyword arguments.
        - args_limit positive integer: Take the args_limit first argument names
            (of signature) as args, and the rest as kwargs.

    >>> def foo(w, /, x: float, y=1, *, z: int = 1):
    ...     return ((w + x) * y) ** z
    >>> foo_sig = Sig(foo)
    >>> args, kwargs = foo_sig.args_and_kwargs_from_kwargs(
    ...     dict(w=4, x=3, y=2, z=1)
    ... )
    >>> assert (args, kwargs) == ((4,), {"x": 3, "y": 2, "z": 1})
    >>> assert foo(*args, **kwargs) == foo(4, 3, 2, z=1) == 14

    The `args_limit` begs explanation.
    Consider the signature of `def foo(w, /, x: float, y=1, *, z: int = 1): ...`
    for instance. We could call the function with the following (args, kwargs) pairs:
    - ((1,), {'x': 2, 'y': 3, 'z': 4})
    - ((1, 2), {'y': 3, 'z': 4})
    - ((1, 2, 3), {'z': 4})
    The two other combinations (empty args or empty kwargs) are not valid
    because of the / and * constraints.

    But when asked for an (args, kwargs) pair, which of the three valid options
    should be returned? This is what the `args_limit` argument controls.

    If `args_limit == 0`, the least args (positional arguments) will be returned.
    It's the default.

    >>> kwargs = dict(w=4, x=3, y=2, z=1)
    >>> foo_sig.args_and_kwargs_from_kwargs(kwargs, args_limit=0)
    ((4,), {'x': 3, 'y': 2, 'z': 1})

    If `args_limit is None`, the least kwargs (keyword arguments) will be returned.

    >>> foo_sig.args_and_kwargs_from_kwargs(kwargs, args_limit=None)
    ((4, 3, 2), {'z': 1})

    If `args_limit` is a positive integer, the first `args_limit` arguments
    will be returned (not checking at all if this is valid!).

    >>> foo_sig.args_and_kwargs_from_kwargs(kwargs, args_limit=1)
    ((4,), {'x': 3, 'y': 2, 'z': 1})
    >>> foo_sig.args_and_kwargs_from_kwargs(kwargs, args_limit=2)
    ((4, 3), {'y': 2, 'z': 1})
    >>> foo_sig.args_and_kwargs_from_kwargs(kwargs, args_limit=3)
    ((4, 3, 2), {'z': 1})

    Note that 'args_limit''s behavior is consistent with list behvior in the sense
    that:

    >>> args = (0, 1, 2, 3)
    >>> args[:0]
    ()
    >>> args[:None]
    (0, 1, 2, 3)
    >>> args[2]
    2

    By default, only the arguments that were given in the kwargs input will be
    returned in the (args, kwargs) output.
    If you also want to get those that have defaults (according to signature),
    you need to specify it with the `apply_defaults=True` argument.

    >>> foo_sig.args_and_kwargs_from_kwargs(dict(w=4, x=3))
    ((4,), {'x': 3})
    >>> foo_sig.args_and_kwargs_from_kwargs(dict(w=4, x=3), apply_defaults=True)
    ((4,), {'x': 3, 'y': 1, 'z': 1})

    By default, all required arguments must be given.
    Not doing so will lead to a `TypeError`.
    If you want to process your arguments anyway, specify `allow_partial=True`.

    >>> foo_sig.args_and_kwargs_from_kwargs(dict(w=4))
    Traceback (most recent call last):
      ...
    TypeError: missing a required argument: 'x'
    >>> foo_sig.args_and_kwargs_from_kwargs(dict(w=4), allow_partial=True)
    ((4,), {})

    Specifying argument names that are not recognized by the signature will
    lead to a `TypeError`.
    If you want to avoid this (and just take from the input `kwargs` what ever you
    can), specify this with `allow_excess=True`.

    >>> foo_sig.args_and_kwargs_from_kwargs(dict(w=4, x=3, extra='stuff'))
    Traceback (most recent call last):
        ...
    TypeError: Got unexpected keyword arguments: extra
    >>> foo_sig.args_and_kwargs_from_kwargs(dict(w=4, x=3, extra='stuff'),
    ...     allow_excess=True)
    ((4,), {'x': 3})

    An edge case: When a `VAR_POSITIONAL` follows a `POSITION_OR_KEYWORD`...

    >>> Sig(lambda a, *b, c=2: None).args_and_kwargs_from_kwargs(
    ...     {"a": 1, "b": [2, 3], "c": 4}
    ... )
    ((1, [2, 3]), {'c': 4})

    See `kwargs_from_args_and_kwargs` (namely for the description of the arguments.
    """

    if args_limit is None:
        # Take the maximum number of args (positional arguments).
        # The only kwargs (keyword arguments) you should have are keyword-only
        # and var-keyword arguments.
        idx = next((i for i, p in enumerate(self.params) if p.kind > VP), None)
        names_for_args = self.names[:idx]
    elif args_limit == 0:
        # Take the minimum number possible of args (positional arguments)
        # Only those that are position only or before a var-positional.
        vp_idx = self.index_of_var_positional
        if vp_idx is None:
            names_for_args = self.names_of_kind[PO]
        else:
            # When there's a VP present, all arguments before it can only be
            # expressed positionally if the VP argument is non-empty.
            # So, here we just consider all arguments positionally up to the VP arg.
            names_for_args = self.names[: (vp_idx + 1)]
    else:
        names_for_args = self.names[:args_limit]

    args = tuple(kwargs[name] for name in names_for_args if name in kwargs)
    kwargs = {name: kwargs[name] for name in kwargs if name not in names_for_args}

    kwargs = self.kwargs_from_args_and_kwargs(
        args,
        kwargs,
        apply_defaults=apply_defaults,
        allow_partial=allow_partial,
        allow_excess=allow_excess,
        ignore_kind=ignore_kind,
        debug=True,
    )
    kwargs = {name: kwargs[name] for name in kwargs if name not in names_for_args}

    return args, kwargs


if __name__ == '__main__':

    # def f(a, **kwargs):
    #    print(kwargs)
    #    return a

    # f_node = FuncNode(func=f)
    # d = DAG([f_node])  # breakpoint
    # args = ['a'=1]
    # kwargs = {'x'=3}

    # d(*args, **kwargs
    # scope = d.sig.kwargs_from_args_and_kwargs((), {"a": 1, "x": 3})
    # assert scope == {"a": 1, "kwargs": {"x": 3}}
    # d.call_on_scope(scope)  # breakpoint
    # assert scope == {"a": 1, "kwargs": {"x": 3}, "f": 1}
    # res = d.extract_output_from_scope(scope, d.leafs)

    # print(res)  # breakpoint
    from i2 import Sig
    from i2.signatures import VK

    def foo(w, /, x: float, y='YY', *, z: str = 'ZZ', **rest):
        pass

    # sig = Sig(foo)
    # res = kwargs_from_args_and_kwargs(
    #    sig, (11, 22, "you"), dict(z="zoo", other="stuff"), debug=True
    # )
    # print(res)
    # assert res == {"w": 11, "x": 22, "y": "you", "z": "zoo", "other": "stuff"}
    # do the reverse one args_from_etc....
    def foo(w, /, x: float, y=1, *args, z: int = 1, **rest):
        return ((w + x) * y) ** z

    foo_sig = Sig(foo)
    args, kwargs = foo_sig.args_and_kwargs_from_kwargs(dict(w=4, x=3, y=2, z=1, t=12))
    print(f'args:{args}, kwargs: {kwargs}')



# TODO: Have all arguments annotated with Callable default to sourcing themselves in
#  all callable elements of the mall (i.e. a aggregate store that only has the
#  callable objects)


from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SelectBox
from typing import Any
from streamlitfront.elements import MultiSourceInput
from typing import Callable
import streamlit as st
from front.elements import InputBase
from i2 import Sig, Pipe
from dataclasses import dataclass

from streamlitfront.elements import TextInput, SelectBox, ExecSection, TextOutput
from functools import partial
from front.crude import Crudifier
from streamlitfront.data_binding import BoundData


def identity(x):
    return x


def three_element_pipe(
    first: Callable = identity,
    middle: Callable = identity,
    last: Callable = identity,
    __name__: str = '',
    __doc__: str = '',
):
    """
    The only purpose for this function is to enable Pipe instances to be
    created in front since we don't have *args handling yet (well... anymore).
    """
    return Pipe(first, middle, last, __name__=__name__, __doc__=__doc__)


from slang import fixed_step_chunker, mk_chk_fft
from slang.spectrop import logarithmic_bands_matrix
from dol import Files, wrap_kvs, filt_iter


# # ============ MALL ===============
# if not b.mall():
#     b.mall = dict(func_store={})

# mall = b.mall()


def foo(x: str, y: str, z: int):
    return x + y * z


def bar(msg: str):
    return msg


# Choice #1: Simple ####################
# data = ['foo', 'bar']
# metadata = {'foo': foo, 'bar': bar}

# Choice #2: Involving a local file store ####################
from dol import Files

data = Files('~', max_levels=0)
metadata = {k: [foo, bar][len(k) % 2] for k in data}

# print(metadata)

# @Crudifier(output_store="func_store", mall=mall)
def my_map(func, kwargs):
    f = metadata[func]
    return partial(f, **kwargs)
    # return f, kwargs


def expand_names(names):
    return {name: {ELEMENT_KEY: TextInput} for name in names}


def populate_kwargs():
    value = b.selected_func()
    list_names = list(Sig(metadata[value]).names)
    b.names_for_kwargs = expand_names(list_names)
    st.write(b.names_for_kwargs())


if not b.selected_func():
    b.selected_func = next(iter(data))


def get_kwargs(**kwargs):
    return kwargs


@dataclass
class KwargsInput(InputBase):
    func_sig: Sig = None

    def __post_init__(self):
        super().__post_init__()
        self.get_kwargs = self.func_sig(get_kwargs)
        self.inputs = self._build_inputs_from_sig()

    def render(self):
        exec_section = ExecSection(
            obj=self.get_kwargs,
            inputs=self.inputs,
            output={ELEMENT_KEY: TextOutput},
            auto_submit=True,
            on_submit=self._return_kwargs,
            use_expander=False,
        )
        exec_section()
        return self.value()

    def _build_inputs_from_sig(self):
        return {
            name: {ELEMENT_KEY: TextInput, 'bound_data_factory': BoundData}
            for name in self.func_sig
        }

    def _return_kwargs(self, output):
        self.value.set(output)


if __name__ == '__main__':
    app = mk_app(
        [my_map],
        config={
            APP_KEY: {'title': 'Rendering map'},
            RENDERING_KEY: {
                'my_map': {
                    'execution': {
                        'inputs': {
                            'func': {
                                ELEMENT_KEY: SelectBox,
                                'options': data,
                                'value': b.selected_func,
                                'on_value_change': populate_kwargs,
                            },
                            'kwargs': {
                                ELEMENT_KEY: KwargsInput,
                                'func_sig': Sig(metadata[b.selected_func()]),
                            },
                        }
                    }
                },
            },
        },
    )
    app()



from dataclasses import dataclass
from i2.wrapper import Wrap
from i2 import Sig
import typing


def foo(self, y):
    return self.x + y


@dataclass
class A:
    x: int
    bar = Wrap(foo)


# want this to be equivalent to:
@dataclass
class B:
    x: int

    def bar(self, y):
        return self.x + y


@dataclass
class B2:
    x: int

    def bar(self, y):
        return self.x + y

    bar2 = bar


def post(wrapper, klass):
    return wrapper.__get__(klass, None)


"""@dataclass
class AA:
    x: int
    bar = post(Wrap(foo), AA)
"""


@dataclass
class AAA:
    x: int
    bar: typing.Callable[..., typing.Any] = foo


class C:
    def __init__(self, x):
        self.x = x
        self.bar = foo


class D:
    def __init__(self, x):
        self.x = x

    bar = foo

    def g(self):
        return 'hello world'

    h = g


if __name__ == '__main__':
    a = A(x=1)
    b = B(x=1)
    c = C(x=1)
    d = D(x=1)
    # aa = AA(x=1)
    aaa = AAA(x=1)
    aaa.bar(3)
    d.bar(3)
    print(type(a.bar))
    print(D(x=1).bar(3))
    # A.bar
    # B.bar
    # b.bar()
    # (method) bar:(y)-> Any
    a.bar(3)
    b2 = B2(x=1)
    b2.bar(3)
    b2.bar2(3)
    # (variable) bar: MethodType
    a.bar.func
    c.bar()



import streamlit as st
import websockets
import asyncio
import base64
import json
import pyaudio
import os
from pathlib import Path

# Session state
if 'text' not in st.session_state:
    st.session_state['text'] = 'Listening...'
    st.session_state['run'] = False

# Audio parameters
st.sidebar.header('Audio Parameters')

FRAMES_PER_BUFFER = int(st.sidebar.text_input('Frames per buffer', 3200))
FORMAT = pyaudio.paInt16
CHANNELS = 1
RATE = int(st.sidebar.text_input('Rate', 16000))
p = pyaudio.PyAudio()

# Open an audio stream with above parameter settings
stream = p.open(
    format=FORMAT,
    channels=CHANNELS,
    rate=RATE,
    input=True,
    frames_per_buffer=FRAMES_PER_BUFFER,
)

# Start/stop audio transmission
def start_listening():
    st.session_state['run'] = True


def download_transcription():
    read_txt = open('transcription.txt', 'r')
    st.download_button(
        label='Download transcription',
        data=read_txt,
        file_name='transcription_output.txt',
        mime='text/plain',
    )


def stop_listening():
    st.session_state['run'] = False


# Web user interface
st.title('🎙️ Real-Time Transcription App')

with st.expander('About this App'):
    st.markdown(
        '''
	This Streamlit app uses the AssemblyAI API to perform real-time transcription.
	
	Libraries used:
	- `streamlit` - web framework
	- `pyaudio` - a Python library providing bindings to [PortAudio](http://www.portaudio.com/) (cross-platform audio processing library)
	- `websockets` - allows interaction with the API
	- `asyncio` - allows concurrent input/output processing
	- `base64` - encode/decode audio data
	- `json` - allows reading of AssemblyAI audio output in JSON format
	'''
    )

col1, col2 = st.columns(2)

col1.button('Start', on_click=start_listening)
col2.button('Stop', on_click=stop_listening)

# Send audio (Input) / Receive transcription (Output)
async def send_receive():
    URL = f'wss://api.assemblyai.com/v2/realtime/ws?sample_rate={RATE}'

    print(f'Connecting websocket to url ${URL}')

    async with websockets.connect(
        URL,
        extra_headers=(('Authorization', st.secrets['api_key']),),
        ping_interval=5,
        ping_timeout=20,
    ) as _ws:

        r = await asyncio.sleep(0.1)
        print('Receiving messages ...')

        session_begins = await _ws.recv()
        print(session_begins)
        print('Sending messages ...')

        async def send():
            while st.session_state['run']:
                try:
                    data = stream.read(FRAMES_PER_BUFFER)
                    data = base64.b64encode(data).decode('utf-8')
                    json_data = json.dumps({'audio_data': str(data)})
                    r = await _ws.send(json_data)

                except websockets.exceptions.ConnectionClosedError as e:
                    print(e)
                    assert e.code == 4008
                    break

                except Exception as e:
                    print(e)
                    assert False, 'Not a websocket 4008 error'

                r = await asyncio.sleep(0.01)

        async def receive():
            while st.session_state['run']:
                try:
                    result_str = await _ws.recv()
                    result = json.loads(result_str)['text']

                    if json.loads(result_str)['message_type'] == 'FinalTranscript':
                        print(result)
                        st.session_state['text'] = result
                        st.write(st.session_state['text'])

                        transcription_txt = open('transcription.txt', 'a')
                        transcription_txt.write(st.session_state['text'])
                        transcription_txt.write(' ')
                        transcription_txt.close()

                except websockets.exceptions.ConnectionClosedError as e:
                    print(e)
                    assert e.code == 4008
                    break

                except Exception as e:
                    print(e)
                    assert False, 'Not a websocket 4008 error'

        send_result, receive_result = await asyncio.gather(send(), receive())


asyncio.run(send_receive())

if Path('transcription.txt').is_file():
    st.markdown('### Download')
    download_transcription()
    os.remove('transcription.txt')



import dash
from dash.dependencies import Output, Input
import dash_core_components as dcc
import dash_html_components as html
import plotly
import plotly.graph_objs as go
import numpy as np

# starting values with 11 data points
X = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
Y = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

# define the app
app = dash.Dash(__name__)

# define the first html block
app.layout = html.Div(
    [
        # contain a graph with a periodic call back thanks to interval
        dcc.Graph(id='live-graph', animate=True),
        dcc.Interval(id='graph-update', interval=1000, n_intervals=0),
    ]
)

# the callback updates the graph
@app.callback(Output('live-graph', 'figure'), [Input('graph-update', 'n_intervals')])

# the function to update the graph
def update_graph(n):

    # add 1 to the x axis
    X.append(X[-1] + 1)

    # the next y value is a random normal
    Y.append(np.random.normal(0, 2))

    # define the data for the plot
    data = plotly.graph_objs.Scatter(
        # always plot the last 10 values of x and y
        x=list(X[-10:]),
        y=list(Y[-10:]),
        name='Scatter',
        mode='lines+markers',
    )

    # recompute the x range based on the last 10 points
    x_range = [min(X[-10:]), max(X[-10:])]

    # recompute y range for overall min and max of y
    y_range = [min(Y), max(Y)]

    # return the plot dictionary
    return {
        'data': [data],
        'layout': go.Layout(xaxis=dict(range=x_range), yaxis=dict(range=y_range),),
    }


if __name__ == '__main__':
    app.run_server()



from collections.abc import Callable
from dataclasses import dataclass
from io import BytesIO
from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
import soundfile as sf
import streamlit as st
from front import APP_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY, RENDERING_KEY
from front.crude import Crudifier, prepare_for_crude_dispatch
from front.elements import FileUploaderBase, OutputBase
from meshed import DAG, code_to_dag
from odat.mdat.vacuum import (
    DFLT_ANNOTS_COLS,
    DFLT_CHUNKER,
    DFLT_FEATURIZER,
    DFLT_LOCAL_SOURCE_DIR,
    Dacc,
    annot_columns,
)
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import normalize
from streamlitfront import mk_app
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    FloatSliderInput,
    MultiSourceInput,
    SelectBox,
    TextInput,
    TextOutput,
)
from streamlitfront.examples.util import Graph
from streamlitfront import binder as b

chunker = DFLT_CHUNKER
featurizer = DFLT_FEATURIZER
learner = RandomForestClassifier(max_depth=2, random_state=0)
learner
dflt_wfsrc = DFLT_LOCAL_SOURCE_DIR


@dataclass
class NumpyArrayDisplay(OutputBase):
    def render(self):
        arr = self.output
        st.write(arr)


def mk_Xy(wfs, annots):
    X_train, y_train, X_test, y_test = [], [], [], []
    for sref, tag, train_info, _ in annots:

        signal = wfs[sref]
        wf = normalize(np.float32(signal).reshape(1, -1))[0]
        for chk in chunker(wf):

            if train_info == 'train':
                X_train.append(featurizer(chk))
                y_train.append(tag)
            if train_info == 'test':
                X_test.append(featurizer(chk))
                y_test.append(tag)

    return np.array(X_train), y_train, np.array(X_test), y_test


def preprocess(X_train):

    from shaded.chained_spectral_projector import learn_chain_proj_matrix

    proj_matrix = learn_chain_proj_matrix(X_train)
    X_train_proj = np.dot(X_train, proj_matrix)

    return proj_matrix, X_train_proj


def train(learner, X_train, y_train):
    proj_matrix, X_train = preprocess(X_train)
    fitted_learner = learner.fit(X_train, y_train)

    return fitted_learner, proj_matrix


def apply(model, preprocessor, X):
    X = np.dot(X, preprocessor)
    return model.predict(X)


def get_wfs(wf_src):
    dacc = Dacc(wf_src)
    return dacc.wfs


def get_annots(wf_store):
    srefs = wf_store.keys()
    annots = annot_columns(srefs)
    return annots


def disp_results(results):
    return len(results)


# func_source: mapping
d = {
    'get_wfs': get_wfs,
    'get_annots': get_annots,
    'mk_Xy': mk_Xy,
    'train': train,
    'apply': apply,
}


if not b.mall():
    b.mall = dict(
        learner={'rtree': RandomForestClassifier(max_depth=2, random_state=0)},
        wfsource={'wf_src': dflt_wfsrc},
        output={},
    )

mall = b.mall()


@Crudifier(
    mall=mall,
    param_to_mall_map={'learner': 'learner', 'wf_src': 'wfsource'},
    output_store='output',
)
def classify(wf_src, learner):
    wf_store = get_wfs(wf_src)
    annots = get_annots(wf_store)
    X_train, y_train, X_test, y_test = mk_Xy(wf_store, annots)
    model, preprocessor = train(learner, X_train, y_train)
    results = apply(model, preprocessor, X_test)
    return results


config_ = {
    APP_KEY: {'title': 'Crudified learner'},
    RENDERING_KEY: {
        'classify': {
            'execution': {
                'inputs': {
                    'wf_src': {ELEMENT_KEY: SelectBox, 'options': mall['wfsource'],},
                    'learner': {ELEMENT_KEY: SelectBox, 'options': mall['learner'],},
                },
                'output': {ELEMENT_KEY: NumpyArrayDisplay,},
            }
        },
    },
}

app = mk_app([classify], config=config_,)
app()



from front.crude import prepare_for_crude_dispatch
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, OBJ_KEY, NAME_KEY
from collections.abc import Callable

from streamlitfront.base import mk_app
from streamlitfront.elements.elements import SelectBox, TextSection
from streamlitfront.examples.util import get_code_of_current_file
import streamlit as st


def foo(a, b: float, c: int):
    """This is foo. It computes something"""
    return a + b * c


param_to_mall_map = dict(a='a', b='b_store', c='c_store')
mall = dict(
    a={'one': 1, 'two': 2},
    b_store={'three': 3, 'four': 4},
    unused_store={'to': 'illustrate'},
    c_store={'value': 4},
)


def crudify(funcs):
    for func in funcs:
        yield prepare_for_crude_dispatch(
            func, param_to_mall_map=param_to_mall_map, mall=mall
        )


app = mk_app(
    [foo],
    config={
        APP_KEY: {'title': 'Crude App'},
        OBJ_KEY: {'trans': crudify},
        RENDERING_KEY: {
            'foo': {
                'execution': {
                    'inputs': {
                        'a': {'options': mall['a'],},
                        'b': {'options': mall['b_store'],},
                        'c': {'options': mall['c_store'],},
                        str: {ELEMENT_KEY: SelectBox,},
                    }
                },
                'code': {
                    ELEMENT_KEY: TextSection,
                    NAME_KEY: 'Source Code',
                    'kind': 'code',
                    'language': 'python',
                    'content': get_code_of_current_file(),
                },
            }
        },
    },
)
app()
st.write(mall)



from typing import Literal, Optional, Callable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, AppMaker
from front.util import deep_merge
from i2 import name_of_obj

from streamlitfront.elements import FloatSliderInput
from streamlitfront import mk_app
from streamlitfront.spec_maker import SpecMaker
from streamlitfront.tools import (
    OBJECT,
    NAME_OF_OBJ,
    SUBTYPE,
    first_element_matching_type,
)

from operator import attrgetter
from functools import partial


def foo(x: int = 100, p: float = 0.5):
    return x * p


def bar(x: int):
    return x ** 2


foo.front_spec = {'execution': {'inputs': {'p': {ELEMENT_KEY: FloatSliderInput,}},}}

from front.elements import ElementTreeMaker
from itertools import chain


def deep_merge_list(obj, updates):
    for update in updates:
        obj = deep_merge(obj, update)
    return obj


class MyAppMaker(AppMaker):
    def __init__(
        self,
        rules,
        spec_maker_factory: Callable,
        element_tree_maker_factory: Callable = ElementTreeMaker,
    ):
        super().__init__(
            spec_maker_factory=spec_maker_factory,
            element_tree_maker_factory=element_tree_maker_factory,
        )
        self.rules = rules

    def mk_app(self, objs, config=None, convention=None):
        config = config or {}
        config_updates = chain.from_iterable(
            mk_find_render_keys(obj, self.rules) for obj in objs
        )
        config_updates = list(config_updates)
        # print(config_updates)
        config = deep_merge_list(config, config_updates)

        return super().mk_app(objs, config=config, convention=convention)


class OriginalMyAppMaker(AppMaker):
    def mk_app(self, objs, config=None, convention=None):
        config = config or {}
        for obj in objs:
            if hasattr(obj, 'front_spec'):
                obj_config = {
                    RENDERING_KEY: {name_of_obj(obj): obj.front_spec},
                }
                config = deep_merge(config, obj_config)
        return super().mk_app(objs, config=config, convention=convention)


# Valentin: modif made to avoid changing too much right now


# TODO: Extract routing logic (the if/elifs) and expose control to interface
def _find_render_keys(objs, render_keys: dict):
    """Make a static configs that has a one to one relationship with objects"""
    # TODO: raise specific error with more info instead of assert:
    assert set(objs) == set(objs), f"Some of your objects weren't unique"

    for obj in objs:
        if obj in render_keys:
            yield OBJECT, (obj, render_keys[obj])
        elif name_of_obj(obj) in render_keys:
            yield NAME_OF_OBJ, (name_of_obj(obj), render_keys[name_of_obj(obj)])
        elif (
            type_ := first_element_matching_type(
                SUBTYPE, (obj, filter(lambda x: isinstance(x, type), render_keys))
            )
        ) is not None:
            yield type_, name_of_obj(type_)
        else:
            # TODO: More significant error (perhaps list ALL the objects that are not
            #  mapped and suggest what to do about it (given the specific objects and
            #  render keys).
            raise ValueError(f"Object couldn't be mapped to a render_keys key: {obj}")


from typing import TypeVar, Tuple, Iterable, Callable

Obj = TypeVar('Obj')
Output = TypeVar('Output')
Cond = Callable[[Obj], bool]
Then = Callable[[Obj], Output]
Rule = Tuple[Cond, Then]
Rules = Iterable[Rule]

# Note: We could iterate over objs or over rules. Context tells what's best.
# typical use is to define rules and apply to objs like so

# >>> from functools import partial
# >>> singular_find_render_keys = partial(mk_find_render_keys, rules=rules)  # fix rules
# >>> find_render_keys = partial(map, singular_find_render_keys)  # apply to iterable


def mk_find_render_keys(obj: Obj, rules: Rules) -> Iterable[Output]:
    for cond, then in rules:
        if cond(obj):
            yield then(obj)


def alt_mk_find_render_keys(obj: Obj, rules: Rules) -> Iterable[Output]:
    for rule in rules:
        if rule.cond(obj):
            yield rule.then(obj)


def has_metadata(obj: Obj, data='front_spec') -> bool:
    return hasattr(obj, data)


DFLT_COND = partial(has_metadata, data='front_spec')
DFLT_RETRIEVE_SPEC = attrgetter('front_spec')


def config_from_attribute(obj: Obj, retrieve_spec: Callable = DFLT_RETRIEVE_SPEC):
    spec_for_obj = retrieve_spec(obj)
    obj_config = {
        RENDERING_KEY: {name_of_obj(obj): spec_for_obj},
    }

    return obj_config


if __name__ == '__main__':
    rules = [(DFLT_COND, config_from_attribute)]

    app_maker = MyAppMaker(spec_maker_factory=SpecMaker, rules=rules)
    app = app_maker.mk_app([foo, bar], config={APP_KEY: {'title': 'Custom App Maker'}})
    app()



from dataclasses import dataclass
from functools import partial
from io import BytesIO
import os
from pathlib import Path
import time
from typing import Any, Iterable, Union
from dol import Files
from front.elements import DEFAULT_INPUT_KEY, OutputBase
from meshed import code_to_dag, DAG
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY
from collections.abc import Callable
from front.crude import Crudifier, prepare_for_crude_dispatch
from streamlitfront.elements import TextInput, SelectBox
from dol.appendable import appendable
import soundfile as sf
import matplotlib.pyplot as plt
from streamlit.uploaded_file_manager import UploadedFile

from streamlitfront import mk_app, binder as b
from streamlitfront.examples.util import Graph
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
import streamlit as st

import os
import soundfile as sf
import pandas as pd
import zipfile
from io import BytesIO
from pathlib import Path

from dol.appendable import add_append_functionality_to_store_cls
from dol import Store
from dol import FilesOfZip, wrap_kvs, filt_iter

from py2store import FilesOfZip
from hear import WavLocalFileStore
from dol import FuncReader


# ============ BACKEND ============
WaveForm = Any
DFLT_WF_PATH = '/Users/sylvain/Dropbox/Otosense/VacuumEdgeImpulse/'
DFLT_ANNOT_PATH = '/Users/sylvain/Dropbox/sipyb/Testing/data/annots_vacuum.csv'


from hear import WavLocalFileStore
from dol import FuncReader


def my_obj_of_data(b):
    return sf.read(BytesIO(b), dtype='float32')[0]


@wrap_kvs(obj_of_data=my_obj_of_data)
@filt_iter(filt=lambda x: not x.startswith('__MACOSX') and x.endswith('.wav'))
class WfZipStore(FilesOfZip):
    """Waveform access. Keys are .wav filenames and values are numpy arrays of int16 waveform."""

    pass


def key_to_ext(k):
    _, ext = os.path.splitext(k)
    if ext.startswith('.'):
        ext = ext[1:]
    return ext


def processor_from_ext(ext):
    if ext.startswith('.'):
        ext = ext[1:]
    if ext in {'zip'}:
        pass
    elif ext in {'wav'}:
        pass


def is_zip_file(filepath):
    return zipfile.is_zipfile(filepath)


def is_dir(filepath):
    return os.path.isdir(filepath)


def key_maker(name, prefix):
    return f'{prefix}_{name}'


def wf_store_factory(filepath):
    key = key_maker(name=filepath, prefix='wf_store')
    tag = 'wf_store'

    if is_dir(filepath):
        data = WavLocalFileStore(filepath)

    elif is_zip_file(filepath):
        data = WfZipStore(filepath)

    return mk_store_item(key, tag, data)


def annot_store_factory(filepath):
    key = key_maker(name=filepath, prefix='annot_store')
    tag = 'annot_store'

    data = pd.read_csv(filepath)

    return mk_store_item(key, tag, data)


def mk_store_item(key, tag, data):
    return dict(key=key, tag=tag, data=data)


# tagged_wf_store = appendable(Files, item2kv=tagged_timestamped_kv)
if not b.mall():
    b.mall = dict(
        wf_store_factory={'wf_factory': wf_store_factory},
        wf_store_path={'wf_path': DFLT_WF_PATH},
        annot_store_path={'annot_path': DFLT_ANNOT_PATH},
        # wf_store_factory=dict(one=1, two=2),
        annot_store_factory={'annot_factory': annot_store_factory},
        data_store=dict(),
        dummy_store=dict(),
    )
mall = b.mall()
crudifier = partial(Crudifier, mall=mall)


def auto_namer(*, arguments):
    return '_'.join(map(str, arguments.values()))


# @crudifier(output_store="wf_store", auto_namer=auto_namer)
@crudifier(
    param_to_mall_map=dict(factory='wf_store_factory', path='wf_store_path'),
    output_store='data_store',
)
def mk_wf_store(factory: Any, path: str):
    result = factory(path)
    st.write(result)
    return factory(path)


@crudifier(
    param_to_mall_map=dict(factory='annot_store_factory', path='annot_store_path')
)
def mk_annot_store(factory: Any, path: str):
    result = factory(path)
    st.write(result)
    return factory(path)


# ============ END BACKEND ============


# ============ FRONTEND ============


@dataclass
class SuccessNotification(OutputBase):
    message: str = 'Success!'

    def render(self):
        return st.success(self.message)


get_wfstore_description = '''
Make a store containing wav files.
'''

config_ = {
    APP_KEY: {'title': 'Data Prep App'},
    RENDERING_KEY: {
        'mk_wf_store': {
            NAME_KEY: 'Make Wf Store',
            'description': {'content': get_wfstore_description},
            'execution': {
                'inputs': {
                    'factory': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['wf_store_factory'],
                        # "options": dict(a="a_choice"),
                        # "display_label": False,
                    },
                    'path': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['wf_store_path'],
                        # "options": dict(a="a_choice"),
                        # "display_label": False,
                    },
                },
                'output': {
                    ELEMENT_KEY: SuccessNotification,
                    'message': 'The wave store has been made successfully.',
                },
            },
        },
        'mk_annot_store': {
            NAME_KEY: 'Make annot Store',
            # "description": {"content": get_wfstore_description},
            'execution': {
                'inputs': {
                    'factory': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['annot_store_factory'],
                        # "options": dict(a="a_choice"),
                        # "display_label": False,
                    },
                    'path': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['annot_store_path'],
                        # "options": dict(a="a_choice"),
                        # "display_label": False,
                    },
                },
                'output': {
                    ELEMENT_KEY: SuccessNotification,
                    'message': 'The wave store has been made successfully.',
                },
            },
        },
        DAG: {'graph': {ELEMENT_KEY: Graph, NAME_KEY: 'Flow',}},
        Callable: {
            'execution': {
                'inputs': {
                    'save_name': {ELEMENT_KEY: TextInput, NAME_KEY: 'Save output as',}
                }
            }
        },
    },
}
# ============ END FRONTEND ============

if __name__ == '__main__':
    app = mk_app([mk_wf_store, mk_annot_store], config=config_)
    app()



from io import BytesIO
from typing import Callable
import streamlit as st
import soundfile as sf
import matplotlib.pyplot as plt
from extrude import mk_web_app
from front import APP_KEY, RENDERING_KEY, NAME_KEY, ELEMENT_KEY
from front.elements import OutputBase
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
    SuccessNotification,
    TextInput,
)

from plunk.vf.extrude.edge_impulse_like import funcs
from plunk.vf.extrude.edge_impulse_like.api import api_url


class FileReader(FileUploader):
    def render(self):
        uploaded_file = super().render()
        if uploaded_file:
            data = uploaded_file.getvalue()
            # print(data)
            # print(type(data))
            return data


class TaggedAudioPlayer(OutputBase):
    def render(self):
        if self.output:
            # wf = self.output['wf'].encode()
            # wf = self.output
            # tag = self.output['tag']
            wf, tag = self.output
            arr = sf.read(BytesIO(wf), dtype='int16')[0]
            tab1, tab2 = st.tabs(['Audio Player', 'Waveform'])
            with tab1:
                st.audio(wf)
            with tab2:
                label = f'Tag={tag}'
                # label = 'Plot'
                fig, ax = plt.subplots(figsize=(15, 5))
                ax.plot(arr, label=f'Tag={label}')
                ax.legend()
                st.pyplot(fig)


get_data_description = '''
Feed the system with wave forms from wav files or directly \
from your microphone, and tag them.
'''


if __name__ == '__main__':
    app = mk_web_app(
        funcs,
        api_url=api_url,
        config={
            APP_KEY: {'title': 'Edge-Impulse-like App'},
            RENDERING_KEY: {
                'tag_wf': {
                    NAME_KEY: 'Get Data',
                    'description': {'content': get_data_description},
                    'execution': {
                        'inputs': {
                            'wf': {
                                ELEMENT_KEY: MultiSourceInput,
                                NAME_KEY: 'Wave Form',
                                'From a file': {
                                    ELEMENT_KEY: FileReader,
                                    'type': 'wav',
                                    'display_label': False,
                                },
                                'From the microphone': {
                                    ELEMENT_KEY: AudioRecorder,
                                    # "save_dir": rootdir,
                                },
                            },
                            'save_name': {NAME_KEY: 'Save as',},
                        },
                        'output': {
                            ELEMENT_KEY: SuccessNotification,
                            'message': 'The wave form has been tagged successfully.',
                        },
                    },
                },
                'get_tagged_wf': {
                    NAME_KEY: 'Data Explorer',
                    'description': {
                        'content': '''Explore the existing tagged wave forms and play them.'''
                    },
                    'execution': {
                        'output': {ELEMENT_KEY: TaggedAudioPlayer,},
                        'auto_submit': True,
                    },
                },
            },
        },
    )
    app()



import streamlit as st
from front.crude import Crudifier
from typing import Any
from front.spec_maker_base import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
import streamlit as st


from streamlitfront import mk_app, binder as b
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY
from streamlitfront.elements import SelectBox

from front.crude import Crudifier


from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep5 import (
    DFLT_CHUNKER_MAKER,
    DFLT_FEATURIZER_MAKER,
)

# ============ MALL ===============
if not b.mall():
    b.mall = dict(factory_store={}, factory_objects_store={})

mall = b.mall()


mall['factory_store'] = [
    'FixedSizeChunkerMaker',
    'FeaturizerMaker',
]

metadata = {
    'FixedSizeChunkerMaker': {'func': DFLT_CHUNKER_MAKER, 'out': 'chunker'},
    'FeaturizerMaker': {'func': DFLT_FEATURIZER_MAKER, 'out': 'featurizer'},
}

# ============ DFLT FACTORIES =====

# select_factories = Crudifier(output_store="factories", mall=mall)(select_factories)


@Crudifier(output_store='factory_objects_store', mall=mall)
def select_factory(factory):
    return metadata[factory]


# ============ BACKEND ============
def make_functions():
    pass


# ============ END BACKEND ============


# ============ FRONT END   ============


config_ = {
    APP_KEY: {'title': 'Data Prepper'},
    RENDERING_KEY: {
        'select_factory': {
            NAME_KEY: 'Select factories',
            # "description": {"content": get_data_description},
            'execution': {
                'inputs': {
                    'factory': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['factory_store'],
                    },
                },
            },
        },
    },
}
# ============ END FRONTEND ============

if __name__ == '__main__':
    app = mk_app([select_factory], config=config_)
    app()
    st.write(mall)



# streamlit run pydantic_dispatch_enum.py

import streamlit_pydantic as sp
import streamlit as st
from typing_extensions import TypedDict
from pydantic import create_model_from_typeddict
from enum import Enum


mall = {'x': {'xs_1': 3, 'xs_2': 5}, 'fvs': {'fv_1': [1, 2, 3], 'fv_2': [4, 5, 6]}}


def mk_Enum_from_store(store, key):
    choices = store[key].keys()
    choice = Enum(f'DynamicEnum_{key}', {item: item for item in choices})
    return choice


def mk_Input_for_store_keys(store, keys):
    vals = {key: mk_Enum_from_store(store, key) for key in keys}
    choices_dict = TypedDict('Choices', vals)
    Input = create_model_from_typeddict(choices_dict)

    return Input


my_input = mk_Input_for_store_keys(store=mall, keys=['x', 'fvs'])

data = sp.pydantic_input(key='my_form', model=my_input)
if data:
    st.write(data)



"""
An example of how to transform (manually) a function that runs some data through a model
(both too complex to enter directly and explicitly in forms) into a streamlit
dispatchable function that uses a store (a Mapping) to manage complex data.
"""

# This is what we want our "dispatchable" wrapper to look like

# There should be real physical stores for those types (FVs, FittedModel) that need them
from typing import Any

FVs = Any
FittedModel = Any

# ---------------------------------------------------------------------------------------
# The function(ality) we want to dispatch:
def apply_model(fitted_model: FittedModel, fvs: FVs, method='transform'):
    method_func = getattr(fitted_model, method)
    return method_func(list(fvs)).tolist()


# ---------------------------------------------------------------------------------------
# The stores that will be used -- here, all stores are just dictionaries, but the
# contract is with the typing.Mapping (read-only here) interface.
import numpy as np
from sklearn.preprocessing import MinMaxScaler

mall = dict(
    fvs=dict(  # Mapping[FVsKey, FVs]
        train_fvs_1=np.array([[1], [2], [3], [5], [4], [2], [1], [4], [3]]),
        train_fvs_2=np.array([[1], [10], [5], [3], [4]]),
        test_fvs=np.array([[1], [5], [3], [10], [-5]]),
    ),
    fitted_model=dict(  # Mapping[FittedModelKey, FittedModel]
        fitted_model_1=MinMaxScaler().fit(
            [[1], [2], [3], [5], [4], [2], [1], [4], [3]]
        ),
        fitted_model_2=MinMaxScaler().fit([[1], [10], [5], [3], [4]]),
    ),
    model_results=dict(),  # Mapping[ResultKey, Result]
)

# ---------------------------------------------------------------------------------------
# dispatchable function:

from front.crude import auto_key


def prepare_for_dispatch(func):
    """Wrap func into something that is ready for dispatch.
    In real live, apply_model_using_stores will be made automatically by wrapping
    apply_model.
    Here is manual, but we're pretending that this wrapping happens automatically...
    """
    # apply_model_using_stores will be made automatically by wrapping apply_model
    # Here we're just pretending that this wrapping happens automatically
    if func.__name__ == 'apply_model':

        def apply_model_using_stores(
            fitted_model: str,
            fvs: str,
            method: str = 'transform',
            # TODO: Have streamlit populate automatically with auto_key:
            save_name: str = '',
        ):
            # make a name if not given explicitly
            save_name = save_name or auto_key(fitted_model=fitted_model, fvs=fvs)
            # get the inputs
            fitted_model = mall['fitted_model'][fitted_model]
            fvs = mall['fvs'][fvs]
            # compute the function
            result = apply_model(fitted_model, fvs, method=method)
            # store the outputs
            mall['model_results'][save_name] = result

            return result  # or not

        # sanity check: test apply_model_using_stores
        assert list(mall['model_results']) == []
        t = apply_model_using_stores(fitted_model='fitted_model_1', fvs='test_fvs')
        print(list(mall['model_results']))
        assert list(mall['model_results']) == [
            'fitted_model=fitted_model_1,fvs=test_fvs'
        ]
        assert all(t == np.array([[0.0], [1.0], [0.5], [2.25], [-1.5]]))
        mall['model_results'].clear()

        return apply_model_using_stores
    else:
        raise ValueError(f"Can't dispatch {func}")


if __name__ == '__main__':
    from crude.util import ignore_import_problems

    with ignore_import_problems:
        from streamlitfront.base import dispatch_funcs

        funcs = [prepare_for_dispatch(apply_model)]
        app = dispatch_funcs(funcs)
        app()
        # app = dispatch_funcs(funcs, {'style': {'root_dir': os.path.dirname(os.path.realpath(__file__))}})
        # app.run_server(debug=True)
